'''test sftpretty.Connection'''

import pytest

from common import conn, LOCAL, VFS
from pathlib import Path
from sftpretty import (CnOpts, Connection, ConnectionException,
                       SSHException)


def test_connection_with(sftpserver):
    '''connect to a public sftp server'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            assert sftp.listdir() == ['pub', 'read.me']


def test_connection_bad_host():
    '''attempt connection to a non-existing server'''
    knownhosts = Path('~/.ssh/known_hosts').expanduser()
    knownhosts.parent.mkdir(exist_ok=True, mode=0o700)
    knownhosts.touch(exist_ok=True, mode=0o644)
    knownhosts.write_bytes((b'localhost ssh-ed25519 '
                            b'AAAAC3NzaC1lZDI1NTE5AAAAIB0g3SG/bbyysJ7f0kqdoWMX'
                            b'hHxxFR7aLJYNIHO/MtsD'))
    with pytest.raises(ConnectionException):
        cnopts = CnOpts()
        cnopts.hostkeys = None
        sftp = Connection('localhost.home.arpa', cnopts=CnOpts(),
                          password='badpass', username='badhost')
        sftp.listdir()


def test_connection_bad_credentials():
    '''attempt connection to a non-existing server'''
    copts = LOCAL.copy()
    copts['password'] = 'badword'
    del copts['private_key'], copts['private_key_pass']
    with pytest.raises(SSHException):
        with Connection(**copts) as sftp:
            sftp.listdir()


def test_connection_good(sftpserver):
    '''connect to a public sftp server'''
    with sftpserver.serve_content(VFS):
        sftp = Connection(**conn(sftpserver))
        sftp.close()



from functools import wraps
from hashlib import new, sha3_512
from io import BytesIO, IOBase
from pathlib import Path, PureWindowsPath
from stat import S_IMODE
from time import sleep


def _callback(filename, bytes_so_far, bytes_total, logger=None):
    message = (f'Transfer of File: [{filename}] @ '
               f'{100.0 * bytes_so_far / bytes_total:.1f}% '
               f'{bytes_so_far:d}:{bytes_total:d} bytes ')
    if logger:
        logger.info(message)
    else:
        print(message)


def drivedrop(filepath):
    if PureWindowsPath(filepath).drive:
        filepath = Path('/').joinpath(*Path(filepath).parts[1:]).as_posix()

    return filepath


def hash(filename, algorithm=sha3_512(), blocksize=65536):
    '''hash contents of a file, file like object or string

    :param bytesIO,IObase,str filename:
        path to file, file object, or string to process
    :param hashlib.hash algorithm:
        hash object to use as digest algorithm
    :param int blocksize:
        size of chunk to read in avoiding memory exhaustion

    :returns: hexdigest

    :raises: Exception

    '''
    buffer = new(algorithm.name)
    if isinstance(filename, str):
        try:
            with open(filename, 'rb') as filestream:
                for chunk in iter(lambda: filestream.read(blocksize), b''):
                    buffer.update(chunk)
        except FileNotFoundError:
            buffer.update(bytes(filename.encode('utf-8')))
    elif isinstance(filename, BytesIO):
        for chunk in iter(lambda: filestream.read1(blocksize), b''):
            buffer.update(chunk)
    elif isinstance(filename, IOBase):
        for chunk in iter(lambda: filestream.read(blocksize), b''):
            buffer.update(chunk)

    return algorithm.hexdigest()


def localtree(container, localdir, remotedir, recurse=True):
    '''recursively descend local directory mapping the tree to a
    dictionary container.

    :param dict container: dictionary object to save directory tree
            {localdir:
                 [(localdir/sub-directory,
                   remotedir/localdir/sub-directory)],}
        {localdir: [(content path, remotedir/content path)],}
    :param str localdir:
        root of local directory to descend, use '.' to start at
        :attr:`.pwd`
    :param str remotedir:
        root of remote directory to append localdir too
        path
    :param bool recurse: *Default: True*. To recurse or not to recurse
        that is the question

    :returns: None

    :raises: Exception

    '''
    try:
        if localdir.startswith(':', 1) or localdir.startswith('\\'):
            localdir = PureWindowsPath(localdir)
        else:
            localdir = Path(localdir).expanduser().absolute()
        for localpath in Path(localdir).iterdir():
            if localpath.is_dir():
                local = localpath.as_posix()
                remote = Path(remotedir).joinpath(localpath.stem).as_posix()
                if localdir.as_posix() in container.keys():
                    container[localdir.as_posix()].append((local, remote))
                else:
                    container[localdir.as_posix()] = [(local, remote)]
                if recurse:
                    localtree(container, local, remote, recurse=recurse)
    except Exception as err:
        raise err


def retry(exceptions, tries=0, delay=3, backoff=2, silent=False, logger=None):
    '''Exception type based retry decorator for all your problematic functions

    :param Exception exceptions:
        exception(s) to check. May be a tuple of exceptions to check.
        IOError or IOError(errno.ECOMM) or (IOError,) or
        (ValueError, IOError(errno.ECOMM)
    :param int tries:
        number of times to try (not retry) before giving up.
    :param int delay:
        initial delay between retries in seconds.
    :param int backoff:
        backoff multiplier.
    :param bool silent:
        if set then no logging will be attempted.
    :param logging.logger logger:
        logger instance to use. If None, print.

    :returns: wrapped function

    :raises: Exception

    '''
    try:
        len(exceptions)
    except TypeError:
        exceptions = (exceptions,)
    all_exception_types = tuple(set(x if type(x) is type else x.__class__
                                    for x in exceptions))
    exception_types = tuple(x for x in exceptions if type(x) is type)
    exception_instances = tuple(x for x in exceptions if type(x) is not type)

    def wrapper(f):
        if tries in (None, 0):
            message = 'Retry: [DISABLED]'
            if not silent:
                if logger:
                    logger.debug(message)
                else:
                    print(message)

            return f

        @wraps(f)
        def _retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except all_exception_types as e:
                    if (not any(x for x in exception_types
                                if isinstance(e, x)) and
                        not any(x for x in exception_instances
                                if type(x) is type(e) and
                                x.args == e.args)):
                        raise
                    msg = (f'Retry ({mtries:d}/{tries:d}):\n'
                           f'{str(e) if str(e) != "" else repr(e)}\n'
                           f'Retrying in {mdelay} second(s)...')
                    if not silent:
                        if logger:
                            logger.warning(msg)
                        else:
                            print(msg)
                    sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff

            return f(*args, **kwargs)
        return _retry
    return wrapper


def st_mode_to_int(val):
    '''SFTAttributes st_mode returns an stat type that shows more than what
    can be set.  Trim off those bits and convert to an int representation.
    if you want an object that was `chmod 711` to return a value of 711, use
    this function

    :param int val: the value of an st_mode attr returned by SFTPAttributes

    :returns int: integer representation of octal mode

    '''
    return int(str(oct(S_IMODE(val)))[-3:])



'''test sftpretty module'''

from common import conn, LOCAL, tempfile_containing, VFS
from pathlib import Path
from sftpretty import Connection
from stat import S_ISLNK


def test_sftp_client(lsftp):
    '''test for access to the underlying, active sftpclient'''
    with Connection(**LOCAL) as sftp:
        assert 'normalize' in dir(sftp.sftp_client)
        assert 'readlink' in dir(sftp.sftp_client)
    assert 'normalize' in dir(lsftp.sftp_client)
    assert 'readlink' in dir(lsftp.sftp_client)


def test_mkdir_p(lsftp):
    '''test mkdir_p simple, testing 2 things, oh well'''
    rdir = 'foo/bar/baz'
    rdir2 = 'foo/bar'
    assert lsftp.exists(rdir) is False
    lsftp.mkdir_p(rdir)
    is_dir = lsftp.isdir(rdir)
    lsftp.rmdir(rdir)
    lsftp.rmdir(rdir2)
    lsftp.mkdir_p(rdir)
    is_dir_partial = lsftp.isdir(rdir)
    lsftp.rmdir(rdir)
    lsftp.rmdir(rdir2)
    lsftp.rmdir('foo')
    assert is_dir
    assert is_dir_partial


# def test_lexists_symbolic(psftp):
#     '''test .lexists() vs. symbolic link'''
#     rsym = 'readme.sym'
#     assert psftp.lexists(rsym)


def test_symlink(lsftp):
    '''test symlink creation'''
    rdest = Path.home().joinpath('honey-boo-boo')
    with tempfile_containing() as fname:
        lsftp.put(fname)
        lsftp.symlink(fname, rdest.as_posix())
        rslt = lsftp.lstat(rdest.as_posix())
        is_link = S_ISLNK(rslt.st_mode)
        lsftp.remove(rdest.as_posix())
        lsftp.remove(Path(fname).name)
    assert is_link


def test_exists(sftpserver):
    '''test .exists() fuctionality'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            rfile = 'pub/foo2/bar1/bar1.txt'
            rbad = 'pub/foo2/bar1/peek-a-boo.txt'
            assert sftp.exists(rfile)
            assert sftp.exists(rbad) is False
            assert sftp.exists('pub')


def test_lexists(lsftp):
    '''test .lexists() functionality'''
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        lsftp.put(fname)
        rbad = Path.home().joinpath('peek-a-boo.txt')
        assert lsftp.lexists(fname)
        lsftp.remove(base_fname)
        assert lsftp.lexists(rbad.as_posix()) is False



'''test sftpretty.put'''

import pytest

from common import conn, tempfile_containing, VFS
from pathlib import Path
from sftpretty import Connection
from time import sleep
from unittest.mock import Mock


def test_put(lsftp):
    '''test upload to localhost'''
    contents = 'now is the time\nfor all good...'
    with tempfile_containing(contents=contents) as fname:
        base_fname = Path(fname).name
        if base_fname in lsftp.listdir():
            lsftp.remove(base_fname)
        assert base_fname not in lsftp.listdir()
        lsftp.put(fname)
        assert base_fname in lsftp.listdir()
        with tempfile_containing(contents='') as tfile:
            lsftp.get(base_fname, tfile)
            assert open(tfile).read() == contents
        # clean up
        lsftp.remove(base_fname)


def test_put_bad_local(sftpserver):
    '''try to put a non-existing file to a read-only server'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            with tempfile_containing() as fname:
                pass
            # tempfile has been removed
            with pytest.raises(OSError):
                sftp.put(fname)


def test_put_callback(lsftp):
    '''test the callback feature of put'''
    cback = Mock(return_value=None)
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        lsftp.chdir(Path.home().as_posix())
        lsftp.put(fname, callback=cback)
        # clean up
        lsftp.remove(base_fname)
    # verify callback was called
    assert cback.call_count


def test_put_confirm(lsftp):
    '''test the confirm feature of put'''
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        lsftp.chdir(Path.home().as_posix())
        result = lsftp.put(fname)
        # clean up
        lsftp.remove(base_fname)
    # verify that an SFTPAttribute like Path.stat() was returned
    assert result.st_size == 8192
    assert result.st_uid is not None
    assert result.st_gid is not None
    assert result.st_atime
    assert result.st_mtime


# TODO
# def test_put_not_allowed(psftp):
#     '''try to put a file to a read-only server'''
#     with tempfile_containing() as fname:
#         with pytest.raises(IOError):
#             psftp.put(fname)


def test_put_preserve_mtime(lsftp):
    '''test that m_time is preserved from local to remote, when put'''
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        base = Path(fname).stat()
        # with Connection(**LOCAL) as sftp:
        result1 = lsftp.put(fname, preserve_mtime=True)
        sleep(2)
        result2 = lsftp.put(fname, preserve_mtime=True)
        # clean up
        lsftp.remove(base_fname)
    # see if times are modified
    # assert base.st_atime == result1.st_atime
    assert int(base.st_mtime) == result1.st_mtime
    # assert result1.st_atime == result2.st_atime
    assert int(result1.st_mtime) == result2.st_mtime


def test_put_resume(lsftp):
    '''test upload resume feature'''
    with tempfile_containing(contents='resume this...') as fname:
        base = Path(fname).stat()
    with tempfile_containing(contents='resume ') as fname:
        partial = lsftp.put(fname)
        with open(fname, 'ab') as fh:
            fh.write('this...'.encode('utf-8'))
        result = lsftp.put(fname, preserve_mtime=True, resume=True)
    assert base.st_size == result.st_size
    assert partial.st_mtime == result.st_mtime



'''use the cd contextmanager prior to paramiko establishing a directory
location'''

from common import conn, VFS
from pathlib import Path
from sftpretty import Connection


def test_issue_65(sftpserver):
    '''using the .cd() context manager prior to setting a directory
    via chdir causes an error'''
    pubpath = Path('/home/test').joinpath('pub')
    with sftpserver.serve_content(VFS):
        cnn = conn(sftpserver)
        cnn['default_path'] = None
        with Connection(**cnn) as sftp:
            assert sftp.getcwd() is None
            with sftp.cd(pubpath.as_posix()):
                pass

            assert sftp.getcwd() == '/'



'''test sftpretty.putfo'''

from io import BytesIO
from unittest.mock import call, Mock


def test_putfo_callback_fsize(lsftp):
    '''test putfo with callback and file_size'''
    rfile = 'a-test-file'
    buf = (b'I will not buy this record, it is scratched\nMy hovercraft '
           b'is full of eels.')
    fsize = len(buf)
    bwrote = fsize - 3
    flo = BytesIO(buf)
    cback = Mock(return_value=None)
    lsftp.putfo(flo, rfile, file_size=bwrote, callback=cback)
    lsftp.remove(rfile)
    assert cback.call_count
    assert cback.call_args_list == [call(fsize, bwrote)]


def test_putfo_callback(lsftp):
    '''test putfo with callback'''
    rfile = 'a-test-file'
    buf = (b'I will not buy this record, it is scratched\nMy hovercraft '
           b'is full of eels.')
    fsize = len(buf)
    flo = BytesIO(buf)
    cback = Mock(return_value=None)
    lsftp.putfo(flo, rfile, callback=cback)
    lsftp.remove(rfile)
    assert cback.call_count
    assert cback.call_args_list == [call(fsize, fsize)]


def test_putfo_flo(lsftp):
    '''test putfo in simple form'''
    rfile = 'a-test-file'
    buf = b'I will not buy this record, it is scratched\nMy hovercraft'\
          b' is full of eels.'
    flo = BytesIO(buf)
    assert rfile not in lsftp.listdir()
    rslt = lsftp.putfo(flo, rfile)
    assert rfile in lsftp.listdir()
    lsftp.remove(rfile)
    assert rslt.st_size == len(buf)


def test_putfo_no_remotepath(lsftp):
    '''test putfo uses uuid fallback when not specifying a remotepath'''
    buf = b'I will not buy this record, it is scratched\nMy hovercraft'\
          b' is full of eels.'
    flo = BytesIO(buf)
    rslt = lsftp.putfo(flo)
    assert rslt.st_size == len(buf)


# TODO
# def test_putfo_ro_srv(psftp):
#     '''test error returned from attempting to putfo to a read-only server'''
#     buf = b'I will not buy this record, it is scratched\nMy hovercraft'\
#           b' is full of eels.'
#     flo = BytesIO(buf)
#     with pytest.raises(TypeError):
#         psftp.putfo(flo)



'''session level fixtures'''

import pytest

from common import LOCAL
from sftpretty import CnOpts, Connection


@pytest.fixture(scope='session')
def lsftp(request):
    '''setup a session long connection to the local sftp server'''
    cnopts = CnOpts(knownhosts=None)
    LOCAL['cnopts'] = cnopts
    lsftp = Connection(**LOCAL)
    request.addfinalizer(lsftp.close)
    return lsftp



'''a template for creating tests that display or duplicate issues'''


from common import conn, USER, USER_HOME_PARENT, VFS
from pathlib import Path
from sftpretty import Connection


# this is the preferred test type as it can be run on the CI server and
# requires no configuarion of a real sftp server.  However issues that involve
# authentication and/or authorization currently have to use a real sftp
# server (see test_issue_xx_lsftp)
def test_issue_xx_sftpserver_plugin(sftpserver):
    '''an example showing how to use the sftpserver plugin in a test'''
    testpath = Path('/home/test').joinpath('pub')
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            home = sftp.pwd
            with sftp.cd():
                sftp.chdir('pub')
                assert sftp.pwd == testpath.as_posix()
            assert home == '/home/test'


def test_issue_xx_local_sftpserver(lsftp):
    '''same as test_issue_xx_sftpserver_plugin but written with the local
    sfptserver mechanism, lsftp'''
    home = lsftp.pwd
    testpath = Path(f'{USER_HOME_PARENT}/{USER}').joinpath('pub')
    # starting condition of default directory should be empty, so we need to
    # construct whatever structure we need prior to peforming the test
    lsftp.mkdir('pub')
    with lsftp.cd():
        lsftp.chdir('pub')
        pubdir = lsftp.pwd.endswith(testpath.as_posix())
    homedir = home == lsftp.pwd

    lsftp.rmdir('pub')

    assert pubdir
    assert homedir



'''test sftpretty.getcwd'''

from common import conn, VFS
from pathlib import Path
from sftpretty import Connection


def test_getcwd_none(sftpserver):
    '''test .getcwd with no default_path arg'''
    with sftpserver.serve_content(VFS):
        cnn = conn(sftpserver)
        cnn['default_path'] = None
        with Connection(**cnn) as sftp:
            assert sftp.getcwd() is None


def test_getcwd_default_path(sftpserver):
    '''test .getcwd when using default_path arg'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            assert sftp.getcwd() == '/home/test'


def test_getcwd_after_chdir(sftpserver):
    '''test getcwd after a chdir operation'''
    pubpath = Path('/home/test').joinpath('pub/foo1')
    with sftpserver.serve_content(VFS):
        cnn = conn(sftpserver)
        cnn['default_path'] = None
        with Connection(**cnn) as sftp:
            sftp.chdir(pubpath.as_posix())
            assert sftp.getcwd() == pubpath.as_posix()



'''test sftpretty.rename'''

from common import tempfile_containing
from pathlib import Path


def test_rename(lsftp):
    '''test rename on remote'''
    contents = 'now is the time\nfor all good...'
    with tempfile_containing(contents=contents) as fname:
        base_fname = Path(fname).name
        if base_fname in lsftp.listdir():
            lsftp.remove(base_fname)
        assert base_fname not in lsftp.listdir()
        lsftp.put(fname)
        lsftp.rename(base_fname, 'bob')
        rdirs = lsftp.listdir()
        assert 'bob' in rdirs
        assert base_fname not in rdirs
        lsftp.remove('bob')


# TODO
# def test_rename_ro(psftp):
#     '''test rename on a read-only server'''
#     with pytest.raises(IOError):
#         psftp.rename('readme.txt', 'bob')



'''test sftpretty.chdir'''

import pytest

from common import conn, VFS
from sftpretty import Connection


def test_chdir_bad_dir(sftpserver):
    '''try to chdir() to a non-existing remote dir'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            with pytest.raises(IOError):
                sftp.chdir('i-dont-exist')



'''test CnOpts.config param'''

from common import conn, SKIP_IF_MAC, PASS, USER, USER_HOME, VFS
from pathlib import Path
from sftpretty import CnOpts, Connection


@SKIP_IF_MAC
# Mac-holes can't be trusted to update user passwords without the old one,
# even root
def test_connection_with_config(sftpserver):
    '''connect to a public sftp server using OpenSSH config'''
    config = Path(f'{USER_HOME}/.ssh/config')
    config.parent.mkdir(exist_ok=True, mode=0o700)
    config.touch(exist_ok=True, mode=0o644)
    with sftpserver.serve_content(VFS):
        config.write_bytes(bytes(('Host 127.0.0.1\n\t'
                                  f'Port {sftpserver.port}\n\t'
                                  f'User {USER}').encode('utf-8')))
        cnopts = CnOpts(config=config.as_posix(), knownhosts='sftpserver.pub')
        params = conn(sftpserver)
        params['cnopts'] = cnopts
        params['password'] = PASS
        del params['port']
        del params['private_key']
        del params['private_key_pass']
        del params['username']
        with Connection(**params) as sftp:
            assert sftp.listdir() == ['pub', 'read.me']
    config.unlink()


def test_connection_with_config_alias(sftpserver):
    '''connect to a public sftp server using OpenSSH config alias'''
    config = Path(f'{USER_HOME}/.ssh/config')
    config.parent.mkdir(exist_ok=True, mode=0o700)
    config.touch(exist_ok=True, mode=0o644)
    config.write_bytes(bytes(('Host test\n\t'
                              'Ciphers aes256-ctr,aes192-ctr\n\t'
                              'Hostname 127.0.0.1\n\t'
                              'IdentityFile id_sftpretty\n\t'
                              f'User {USER}').encode('utf-8')))
    cnopts = CnOpts(config=config.as_posix(), knownhosts='sftpserver.pub')
    with sftpserver.serve_content(VFS):
        params = conn(sftpserver)
        params['cnopts'] = cnopts
        params['host'] = 'test'
        del params['private_key']
        del params['username']
        with Connection(**params) as sftp:
            local, remote = sftp.active_ciphers
            assert local in ('aes256-ctr', 'aes192-ctr')
            assert remote in ('aes256-ctr', 'aes192-ctr')
            assert sftp.listdir() == ['pub', 'read.me']
    config.unlink()


def test_connection_with_config_identity(sftpserver):
    '''connect to a public sftp server using an OpenSSH config identity'''
    config = Path(f'{USER_HOME}/.ssh/config')
    config.parent.mkdir(exist_ok=True, mode=0o700)
    config.touch(exist_ok=True, mode=0o644)
    config.write_bytes(bytes(('Host 127.0.0.1\n\t'
                              'IdentityFile id_sftpretty\n\t'
                              f'User {USER}').encode('utf-8')))
    cnopts = CnOpts(config=config.as_posix(), knownhosts='sftpserver.pub')
    with sftpserver.serve_content(VFS):
        params = conn(sftpserver)
        params['cnopts'] = cnopts
        del params['private_key']
        del params['username']
        with Connection(**params) as sftp:
            assert sftp.listdir() == ['pub', 'read.me']
    config.unlink()



'''test sftpretty.isdir and sftpretty.isfile methods'''

from common import conn, VFS
from sftpretty import Connection


def test_isfile(sftpserver):
    '''test .isfile() functionality'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            rfile = 'pub/make.txt'
            rdir = 'pub'
            assert sftp.isfile(rfile)
            assert sftp.isfile(rdir) is False


# TODO
# def test_isfile_2(sftp):
#     '''test .isfile() functionality against a symlink'''
#     rsym = 'readme.sym'
#     assert sftp.isfile(rsym)


def test_isdir(sftpserver):
    '''test .isdir() functionality'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            rfile = 'pub/make.txt'
            rdir = 'pub'
            assert sftp.isdir(rfile) is False
            assert sftp.isdir(rdir)


# TODO
# def test_isdir_2(sftp):
#     '''test .isdir() functionality against a symlink'''
#     rsym = 'readme.sym'
#     assert sftp.isdir(rsym) is False



'''test sftpretty.compression param'''


def test_security_options(lsftp):
    '''test the security_options property has expected attributes and that
    they are tuples'''
    secopts = lsftp.security_options
    for attr in ['ciphers', 'compression', 'digests', 'kex', 'key_types']:
        print(attr)
        assert hasattr(secopts, attr)
        assert isinstance(getattr(secopts, attr), tuple)



'''test sftpretty.get_r'''

from common import conn, rmdir, VFS
from pathlib import Path
from sftpretty import Connection, hash, localtree
from tempfile import mkdtemp


def test_get_r(sftpserver):
    '''test the get_r for remotepath is pwd '.' '''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            localpath = Path(mkdtemp()).as_posix()
            sftp.get_r('.', localpath)

            local_tree = {}
            remote_tree = {}

            remote_cwd = sftp.pwd

            localtree(local_tree, localpath, remote_cwd)
            sftp.remotetree(remote_tree, remote_cwd, localpath)

            localdirs = sorted([localdir.replace(localpath, remote_cwd)
                                for localdir in local_tree.keys()])
            remotedirs = sorted(remote_tree.keys())

            assert localdirs == remotedirs

            rmdir(localpath)


def test_get_r_pwd(sftpserver):
    '''test the get_r for remotepath is pwd '/pub/foo2' '''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            localpath = Path(mkdtemp()).as_posix()
            sftp.get_r('pub/foo2', localpath)

            local_tree = {}
            remote_tree = {}

            remote_cwd = sftp.pwd

            localtree(local_tree, localpath, remote_cwd)
            sftp.remotetree(remote_tree, remote_cwd, localpath)

            localdirs = sorted([localdir.replace(localpath, remote_cwd)
                                for localdir in local_tree.keys()])
            remotedirs = sorted(remote_tree.keys())

            assert localdirs == remotedirs

            rmdir(localpath)


def test_get_r_pathed(sftpserver):
    '''test the get_r for localpath, starting deeper then pwd '''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo2')
            localpath = Path(mkdtemp()).as_posix()
            sftp.get_r('./bar1', localpath)

            local_tree = {}
            remote_tree = {}

            remote_cwd = sftp.pwd

            localtree(local_tree, localpath, remote_cwd)
            sftp.remotetree(remote_tree, remote_cwd, localpath)

            actual = hash(remote_cwd + '/bar1.txt')
            expected = ('a69f73cca23a9ac5c8b567dc185a756e97c982164fe258'
                        '59e0d1dcc1475c80a615b2123af1f5f94c11e3e9402c3a'
                        'c558f500199d95b6d3e301758586281dcd26')

            assert local_tree.keys() == remote_tree.keys()
            assert actual == expected

            rmdir(localpath)


def test_get_r_cdd(sftpserver):
    '''test the get_r for chdir('pub/foo2')'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            localpath = Path(mkdtemp()).as_posix()
            sftp.chdir('pub/foo2')
            sftp.get_r('.', localpath)

            local_tree = {}
            remote_tree = {}

            remote_cwd = sftp.pwd

            localtree(local_tree, localpath, remote_cwd)
            sftp.remotetree(remote_tree, remote_cwd, localpath)

            localdirs = sorted([localdir.replace(localpath, remote_cwd)
                                for localdir in local_tree.keys()])
            remotedirs = sorted(remote_tree.keys())

            assert localdirs == remotedirs

            rmdir(localpath)



'''common setup code for tests'''

import pytest

from contextlib import contextmanager
from os import close, environ
from pathlib import Path
from sftpretty import CnOpts
from tempfile import mkstemp


PASS = 'tEst@!357'
SKIP_IF_CI = pytest.mark.skipif(environ.get('CI', '') > '', reason='Not Local')
SKIP_IF_MAC = pytest.mark.skipif(environ.get('RUNNER_OS', '') == 'macOS',
                                 reason='WhackMac')
SKIP_IF_WIN = pytest.mark.skipif(environ.get('RUNNER_OS', '') == 'Windows',
                                 reason='NoWinZone')
STARS8192 = '*' * 8192
USER = environ.get('USER', environ.get('USERNAME'))
USER_HOME = Path.home().as_posix()
USER_HOME_PARENT = Path(USER_HOME).parent.as_posix()


LOCAL = {'default_path': USER_HOME, 'host': 'localhost',
         'private_key': 'id_sftpretty', 'private_key_pass': PASS,
         'username': USER}


def conn(sftpsrv):
    '''return a dictionary holding argument info for the sftpretty client'''
    cnopts = CnOpts(knownhosts='sftpserver.pub')
    return {'cnopts': cnopts, 'default_path': '/home/test',
            'host': sftpsrv.host, 'port': sftpsrv.port,
            'private_key': 'id_sftpretty', 'private_key_pass': PASS,
            'username': USER}


def rmdir(dir):
    dir = Path(dir)
    for item in dir.iterdir():
        if item.is_dir():
            rmdir(item)
        else:
            item.unlink()
    dir.rmdir()


@contextmanager
def tempfile_containing(contents=STARS8192, suffix=''):
    '''create a temporary file, with optional suffix and return the filename,
    cleanup when finished'''

    fd, temp_path = mkstemp(suffix=suffix)
    close(fd)

    with open(temp_path, 'wb') as fh:
        fh.write(contents.encode('utf-8'))

    try:
        yield Path(temp_path).as_posix()
    finally:
        Path(temp_path).unlink()


# filesystem served by pytest-sftpserver plugin
VFS = {
    'home': {
        'test': {
            'pub': {
                'foo1': {
                    'foo1.txt': 'content of foo1.txt',
                    'image01.jpg': 'data for image01.jpg'
                },
                'make.txt': 'content of make.txt',
                'foo2': {
                    'bar1': {
                        'bar1.txt': 'contents bar1.txt'
                    },
                    'foo2.txt': 'content of foo2.txt'
                }
            },
            'read.me': 'contents of read.me'
        }
    }
}



'''test sftpretty.listdir'''

from common import STARS8192
from io import BytesIO


def test_truncate_smaller(lsftp):
    '''test truncate, make file smaller'''
    flo = BytesIO(bytes(STARS8192, 'UTF-8'))
    rname = 'truncate.txt'
    try:
        lsftp.remove(rname)
    except IOError:
        pass
    lsftp.putfo(flo, rname)
    new_size = lsftp.truncate(rname, 4096)
    assert new_size == 4096
    lsftp.remove(rname)


def test_truncate_larger(lsftp):
    '''test truncate, make file larger'''
    flo = BytesIO(bytes(STARS8192, 'UTF-8'))
    rname = 'truncate.txt'
    try:
        lsftp.remove(rname)
    except IOError:
        pass
    lsftp.putfo(flo, rname)
    new_size = lsftp.truncate(rname, 2 * 8192)
    assert new_size == 2 * 8192
    lsftp.remove(rname)


def test_truncate_same(lsftp):
    '''test truncate, make file same size'''
    flo = BytesIO(bytes(STARS8192, 'UTF-8'))
    rname = 'truncate.txt'
    try:
        lsftp.remove(rname)
    except IOError:
        pass
    lsftp.putfo(flo, rname)
    new_size = lsftp.truncate(rname, 8192)
    assert new_size == 8192
    lsftp.remove(rname)


# TODO
# def test_truncate_ro(psftp):
#     '''test truncate, against read-only server'''
#     rname = Path.home().joinpath('readme.txt').as_posix()
#     with pytest.raises(IOError):
#         _ = psftp.truncate(rname, 8192)



'''test sftpretty.localtree'''

from common import conn, rmdir, VFS
from pathlib import Path
from sftpretty import Connection, localtree
from tempfile import mkdtemp


def test_localtree(sftpserver):
    '''test the localtree function, with recurse'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            localpath = Path(mkdtemp()).as_posix()
            sftp.get_r('.', localpath)

            cwd = sftp.pwd
            tree = {}

            localtree(tree, localpath, cwd)

            local = {
                f'{localpath}': [
                    (f'{localpath}/pub', cwd + '/pub')
                ],
                f'{localpath}/pub': [
                    (f'{localpath}/pub/foo1', cwd + '/pub/foo1'),
                    (f'{localpath}/pub/foo2', cwd + '/pub/foo2')
                ],
                f'{localpath}/pub/foo2': [
                    (f'{localpath}/pub/foo2/bar1', cwd + '/pub/foo2/bar1')
                ]
            }

            for branch in sorted(tree.keys()):
                assert set(local[branch]) == set(tree[branch])
                del tree[branch]
            assert tree == {}

    rmdir(localpath)


def test_localtree_no_recurse(sftpserver):
    '''test the localtree function, without recursing'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            localpath = Path(mkdtemp()).as_posix()
            sftp.chdir('pub/foo2')
            sftp.get_r('.', localpath)

            cwd = sftp.pwd
            tree = {}

            localtree(tree, localpath, cwd, recurse=False)

            local = {
                f'{localpath}': [
                    (f'{localpath}/bar1', cwd + '/bar1')
                ]
            }

            for branch in sorted(tree.keys()):
                assert set(local[branch]) == set(tree[branch])
                del tree[branch]
            assert tree == {}

    rmdir(localpath)



'''test sftpretty.stat and .lstat'''

from blddirs import build_dir_struct
from common import conn, rmdir, VFS
from pathlib import Path
from sftpretty import Connection
from tempfile import mkdtemp


def test_stat(sftpserver):
    '''test stat'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            dirname = 'pub'
            rslt = sftp.stat(dirname)
            assert rslt.st_size >= 0


def test_lstat(lsftp):
    '''test lstat minimal, have to use real server, plugin doesn't support
    lstat'''
    localpath = Path(mkdtemp()).as_posix()
    build_dir_struct(localpath)
    dirname = Path(localpath).joinpath('pub').as_posix()
    rslt = lsftp.lstat(dirname)

    assert rslt.st_size >= 0

    rmdir(localpath)



'''test sftpretty.cd'''

import pytest

from common import conn, VFS
from pathlib import Path
from sftpretty import Connection


def test_cd_none(sftpserver):
    '''test sftpretty.cd with None'''
    pubpath = Path('/home/test').joinpath('pub')
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            home = sftp.pwd
            with sftp.cd():
                sftp.chdir('pub')
                assert sftp.pwd == pubpath.as_posix()
            assert home == pubpath.parent.as_posix()


def test_cd_path(sftpserver):
    '''test sftpretty.cd with a path'''
    pubpath = Path('/home/test').joinpath('pub')
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            home = sftp.pwd
            with sftp.cd('pub'):
                assert sftp.pwd == pubpath.as_posix()
            assert home == pubpath.parent.as_posix()


def test_cd_nested(sftpserver):
    '''test nested cd's'''
    pubpath = Path('/home/test').joinpath('pub')
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            home = sftp.pwd
            with sftp.cd('pub'):
                assert sftp.pwd == pubpath.as_posix()
                with sftp.cd('foo1'):
                    assert sftp.pwd == pubpath.joinpath('foo1').as_posix()
                assert sftp.pwd == pubpath.as_posix()
            assert home == pubpath.parent.as_posix()


def test_cd_bad_path(sftpserver):
    '''test sftpretty.cd with a bad path'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            home = sftp.pwd
            with pytest.raises(IOError):
                with sftp.cd('not-there'):
                    pass
            assert home == '/home/test'



'''test sftpretty.get_d'''

from common import conn, rmdir, VFS
from pathlib import Path
from sftpretty import Connection
from tempfile import mkdtemp


def test_get_d(sftpserver):
    '''test the get_d for remotepath is pwd '.' '''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub')
            localpath = Path(mkdtemp()).as_posix()
            sftp.get_d('.', localpath)

            checks = [(['', ], ['make.txt', ]), ]
            for pth, fls in checks:
                assert sorted([
                    path.name
                    for path in Path(localpath).joinpath(*pth).iterdir()
                ]) == fls

            # cleanup local
            rmdir(localpath)


def test_get_d_pathed(sftpserver):
    '''test the get_d for localpath, starting deeper then pwd '''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub')
            localpath = Path(mkdtemp()).as_posix()
            sftp.get_d('foo1', localpath)

            checks = [(['', ], ['foo1.txt', 'image01.jpg']), ]
            for pth, fls in checks:
                assert sorted([
                    path.name
                    for path in Path(localpath).joinpath(*pth).iterdir()
                ]) == fls

            # cleanup local
            rmdir(localpath)



'''test sftpretty.put_r'''

import pytest

from blddirs import build_dir_struct
from common import rmdir
from pathlib import Path
from tempfile import mkdtemp


def test_put_r(lsftp):
    '''test put_r'''
    localpath = Path(mkdtemp()).as_posix()
    remote = Path.home()
    build_dir_struct(localpath)
    local = Path(localpath).joinpath('pub')
    lsftp.put_r(local.as_posix(), remote.as_posix())

    rmdir(localpath)


# TODO
# def test_put_r_ro(psftp):
#     '''test put_r failure on remote read-only srvr'''
#     # run the op
#     with pytest.raises(IOError):
#         psftp.put_r('.', '.')


def test_put_r_bad_local(lsftp):
    '''test put_r failure on non-existing local directory'''
    # run the op
    with pytest.raises(OSError):
        lsftp.put_r('/non-existing', '.')



'''test sftpretty.chmod'''

import pytest

from common import conn, SKIP_IF_WIN, tempfile_containing, VFS
from pathlib import Path
from sftpretty import Connection
from sftpretty.helpers import st_mode_to_int


def test_chmod_not_exist(sftpserver):
    '''verify error if trying to chmod something that isn't there'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            with pytest.raises(IOError):
                sftp.chmod('i-do-not-exist.txt', 666)


@SKIP_IF_WIN
def test_chmod_simple(lsftp):
    '''test basic chmod with octal mode represented by an int'''
    new_mode = 711
    with tempfile_containing(contents='') as fname:
        base_fname = Path(fname).name
        org_attrs = lsftp.put(fname)
        lsftp.chmod(base_fname, new_mode)
        new_attrs = lsftp.stat(base_fname)
        lsftp.remove(base_fname)

    assert st_mode_to_int(new_attrs.st_mode) == new_mode
    assert new_attrs.st_mode != org_attrs.st_mode


# TODO
# def test_chmod_fail_ro(psftp):
#     '''test chmod against read-only server'''
#     new_mode = 440
#     fname = 'readme.txt'
#     with pytest.raises(IOError):
#         psftp.chmod(fname, new_mode)



'''test sftpretty module, setup.py and tests'''

from pep8 import StyleGuide


def test_pep8():
    '''pep8 check the source'''
    # list the specific files or directories to check, directories are recursed
    paths = ['sftpretty', 'tests']
    p8c = StyleGuide(ignore=['E231'])
    report = p8c.check_files(paths=paths)
    report.print_statistics()
    assert report.get_count() == 0



'''test sftpretty.open'''

from common import VFS, conn
from sftpretty import Connection


def test_open_read(sftpserver):
    '''test the open function'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as psftp:
            psftp.chdir('pub')
            rfile = psftp.open('make.txt')
            contents = rfile.read()
            rfile.close()
            assert contents == b'content of make.txt'


def test_open_read_with(sftpserver):
    '''test the open function in a with statment'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as psftp:
            psftp.chdir('pub')
            with psftp.open('make.txt') as rfile:
                contents = rfile.read()
            assert contents == b'content of make.txt'



'''test sftpretty.mkdir'''

from common import conn, VFS
from sftpretty import Connection
from sftpretty.helpers import st_mode_to_int


def test_mkdir_mode(lsftp):
    '''test mkdir with mode set to 711'''
    dirname = 'test-dir'
    mode = 700
    assert dirname not in lsftp.listdir()
    lsftp.mkdir(dirname, mode=mode)
    attrs = lsftp.stat(dirname)
    lsftp.rmdir(dirname)
    assert st_mode_to_int(attrs.st_mode) == mode


def test_mkdir(sftpserver):
    '''test mkdir'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            dirname = 'test-dir'
            assert dirname not in sftp.listdir()
            sftp.mkdir(dirname)
            assert dirname in sftp.listdir()
            # clean up
            sftp.rmdir(dirname)


# TODO
# def test_mkdir_ro(psftp):
#     '''test mkdir on a read-only server'''
#     dirname = 'test-dir'
#     assert dirname not in psftp.listdir()
#     with pytest.raises(IOError):
#         psftp.mkdir(dirname)



'''test sftpretty.readlink'''

from io import BytesIO
from pathlib import Path


def test_readlink(lsftp):
    '''test the readlink method'''
    buf = b'I will not buy this record, it is scratched\nMy hovercraft'\
          b' is full of eels.'
    flo = BytesIO(buf)
    rfile = 'readme.txt'
    rlink = 'readme.sym'
    rpath = Path.home().joinpath(rfile).as_posix()
    lsftp.putfo(flo, rfile)
    lsftp.symlink(rfile, rlink)

    result = lsftp.readlink(rlink).endswith(rpath)
    lsftp.remove(rlink)
    lsftp.remove(rfile)
    # test assert after cleanup
    assert result



'''test sftpretty.rmdir'''


def test_rmdir(lsftp):
    '''test mkdir'''
    dirname = 'test-rm'
    lsftp.mkdir(dirname)
    assert dirname in lsftp.listdir()
    lsftp.rmdir(dirname)
    assert dirname not in lsftp.listdir()

# TODO
# def test_rmdir_ro(psftp):
#     '''test rmdir against read-only server'''
#     psftp.chdir(Path.home().as_posix())
#     with pytest.raises(IOError):
#         psftp.rmdir('pub')



import pytest

from logging import DEBUG, getLogger, StreamHandler
from sftpretty.helpers import retry


class RetryableError(Exception):
    pass


class AnotherRetryableError(Exception):
    pass


class UnexpectedError(Exception):
    pass


def test_no_retry_required():
    counter = 0

    @retry(RetryableError, tries=4, delay=0.1)
    def succeeds():
        nonlocal counter
        counter += 1
        return 'success'

    r = succeeds()

    assert r == 'success'
    assert counter == 1


def test_retries_once():
    counter = 0

    @retry(RetryableError, tries=4, delay=0.1)
    def fails_once():
        nonlocal counter
        counter += 1
        if counter < 2:
            raise RetryableError('failed')
        else:
            return 'success'

    r = fails_once()

    assert r == 'success'
    assert counter == 2


def test_limit_is_reached():
    counter = 0

    @retry(RetryableError, tries=4, delay=0.1)
    def always_fails():
        nonlocal counter
        counter += 1
        raise RetryableError('failed')

    with pytest.raises(RetryableError, match='failed'):
        always_fails()

    assert counter == 4


def test_multiple_exception_types():
    counter = 0

    @retry((RetryableError, AnotherRetryableError), tries=4, delay=0.1)
    def raise_multiple_exceptions():
        nonlocal counter
        counter += 1
        if counter == 1:
            raise RetryableError('a retryable error')
        elif counter == 2:
            raise AnotherRetryableError('another retryable error')
        else:
            return 'success'

    r = raise_multiple_exceptions()

    assert r == 'success'
    assert counter == 3


def test_unexpected_exception_does_not_retry():

    @retry(RetryableError, tries=4, delay=0.1)
    def raise_unexpected_error():
        raise UnexpectedError('unexpected error')

    with pytest.raises(UnexpectedError, match='unexpected error'):
        raise_unexpected_error()


@pytest.fixture(autouse=True)
def test_using_a_logger(caplog):
    _caplog = caplog
    counter = 0
    expected = {'DEBUG': 'success',
                'ERROR': 'failed',
                'WARNING': ('Retry (4/4):\nfailed\nRetrying in 0.1 '
                            'second(s)...')}
    records = {}

    sh = StreamHandler()
    log = getLogger(__name__)
    log.addHandler(sh)

    @retry(RetryableError, tries=4, delay=0.1, logger=log)
    def fails_once():
        nonlocal counter
        counter += 1
        if counter < 2:
            log.error('failed')
            raise RetryableError('failed')
        else:
            log.debug('success')
            for record in _caplog.records:
                records[record.levelname] = record.message
            return 'success'

    with _caplog.at_level(DEBUG):
        r = fails_once()

    assert r == 'success'
    assert counter == 2
    assert expected == records



'''test sftpretty.timeout'''


def test_timeout_getter(lsftp):
    '''test getting the timeout value'''
    # always starts at no timeout,
    assert lsftp.timeout is None


def test_timeout_setter(lsftp):
    '''test setting the timeout value'''
    lsftp.timeout = 10.5
    assert lsftp.timeout == 10.5
    lsftp.timeout = None
    assert lsftp.timeout is None



'''test CnOpts.log, CnOpts.log_level params and temporary log file creation'''

from common import conn, VFS
from logging import getLogger
from pathlib import Path
from sftpretty import CnOpts, Connection


def test_log_cnopts_explicit_false(sftpserver):
    '''test .logfile returns false when CnOpts.log is set to false'''
    copts = conn(sftpserver)
    cnopts = CnOpts(knownhosts='sftpserver.pub')
    copts['cnopts'] = cnopts
    with sftpserver.serve_content(VFS):
        with Connection(**copts) as sftp:
            assert sftp.logfile is False


def test_log_cnopts_log_level(sftpserver):
    '''test log level is passed to application logger'''
    copts = conn(sftpserver)
    cnopts = CnOpts(knownhosts='sftpserver.pub')
    cnopts.log = True
    cnopts.log_level = 'error'
    copts['cnopts'] = cnopts
    with sftpserver.serve_content(VFS):
        with Connection(**copts) as sftp:
            sftp.listdir()
            log = getLogger('SFTPretty')
            assert log.level == 40


def test_log_cnopts_true(sftpserver):
    '''test .logfile returns temp filename when CnOpts.log is set to True'''
    copts = conn(sftpserver)
    cnopts = CnOpts(knownhosts='sftpserver.pub')
    cnopts.log = True
    copts['cnopts'] = cnopts
    with sftpserver.serve_content(VFS):
        with Connection(**copts) as sftp:
            # we are not writing to a file named 'True'
            assert sftp.logfile == cnopts.log
            assert Path(sftp.logfile).exists()


def test_log_cnopts_user_file(sftpserver):
    '''test .logfile returns temp filename when CnOpts.log is set to True'''
    copts = conn(sftpserver)
    cnopts = CnOpts(knownhosts='sftpserver.pub')
    cnopts.log = Path('~/my-logfile1.txt').expanduser().as_posix()
    copts['cnopts'] = cnopts
    with sftpserver.serve_content(VFS):
        with Connection(**copts) as sftp:
            sftp.listdir()
            assert sftp.logfile == cnopts.log
            assert Path(sftp.logfile).exists()



'''methods to build a known local directory structure, used in testing'''

from common import STARS8192
from pathlib import Path


DIR_LIST = [('pub', ),
            ('pub', 'foo1'),
            ('pub', 'foo2'),
            ('pub', 'foo2', 'bar1')]
FILE_LIST = [('read.me',),
             ('pub', 'make.txt'),
             ('pub', 'foo1', 'foo1.txt'),
             ('pub', 'foo2', 'foo2.txt'),
             ('pub', 'foo2', 'bar1', 'bar1.txt')]


def build_dir_struct(local_path):
    '''build directory structure'''
    for dparts in DIR_LIST:
        Path(local_path).joinpath(*dparts).mkdir()
    for fparts in FILE_LIST:
        with open(Path(local_path).joinpath(*fparts).as_posix(),
                  'wb') as fhndl:
            try:
                fhndl.write(STARS8192)
            except TypeError:
                fhndl.write(bytes(STARS8192, 'UTF-8'))



'''test CnOpts.ciphers param'''

# these can not use fixtures as we need to set ciphers prior to the connection
# being made and fixtures are already active connections.

from common import LOCAL
from sftpretty import CnOpts, Connection


def test_connection_ciphers_cnopts():
    '''test the CnOpts.ciphers portion of the Connection'''
    ciphers = ('aes256-ctr', 'aes256-cbc')
    copts = LOCAL.copy()  # don't sully the module level variable
    cnopts = CnOpts(knownhosts=None)
    cnopts.ciphers = ciphers
    copts['cnopts'] = cnopts
    assert copts != LOCAL
    with Connection(**copts) as sftp:
        rslt = sftp.listdir()
        assert len(rslt) > 1


def test_active_ciphers():
    '''test that method returns a tuple of strings, that show ciphers used'''
    ciphers = ('aes256-ctr', 'aes256-cbc')
    copts = LOCAL.copy()  # don't sully the module level variable
    cnopts = CnOpts(knownhosts=None)
    cnopts.ciphers = ciphers
    copts['cnopts'] = cnopts
    with Connection(**copts) as sftp:
        local_cipher, remote_cipher = sftp.active_ciphers
    assert local_cipher in ciphers
    assert remote_cipher in ciphers



'''test sftpretty.listdir'''

from common import conn, VFS
from sftpretty import Connection


def test_listdir(sftpserver):
    '''test listdir'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as psftp:
            psftp.chdir('pub')
            assert psftp.listdir() == ['foo1', 'foo2', 'make.txt']


def test_listdir_attr(sftpserver):
    '''test listdir'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as psftp:
            psftp.chdir('pub')
            attrs = psftp.listdir_attr()
            assert len(attrs) == 3
            # test they are in filename order
            assert attrs[0].filename == 'foo1'
            assert attrs[1].filename == 'foo2'
            assert attrs[2].filename == 'make.txt'
            # test that longname is there
            for attr in attrs:
                assert attr.longname is not None



'''test sftpretty.remote_server_key'''

import pytest

from common import conn, VFS
from paramiko.hostkeys import HostKeys
from paramiko.rsakey import RSAKey
from pathlib import Path
from sftpretty import CnOpts, Connection, HostKeysException, SSHException


def test_remote_server_key(sftpserver):
    '''test .remote_server_key property'''
    with sftpserver.serve_content(VFS):
        this_conn = conn(sftpserver)
        this_conn['cnopts'].hostkeys = None     # turn-off hostkey verification
        with Connection(**this_conn) as sftp:
            rsk = sftp.remote_server_key
            hks = HostKeys()
            hks.add(hostname=sftpserver.host,
                    keytype=rsk.get_name(),
                    key=rsk)
            hks.save('sftpserver.pub')


def test_cnopts_bad_knownhosts():
    '''test setting knownhosts to a not understood file'''
    with pytest.raises(HostKeysException):
        with pytest.raises(UserWarning):
            knownhosts = Path('~/knownhosts').expanduser().as_posix()
            Path(knownhosts).touch(mode=0o600)
            CnOpts(knownhosts=knownhosts)
            Path(knownhosts).unlink()


def test_cnopts_no_knownhosts():
    '''test setting knownhosts to a non-existant file'''
    with pytest.raises(UserWarning):
        CnOpts(knownhosts='i-m-not-there')


def test_cnopts_none_knownhosts():
    '''test setting knownhosts to None for those with no default known_hosts'''
    knownhosts = Path('~/.ssh/known_hosts').expanduser().as_posix()
    if Path(knownhosts).exists():
        Path(knownhosts).unlink()
    cnopts = CnOpts(knownhosts=None)
    assert cnopts.hostkeys is None


def test_hostkey_not_found():
    '''test that an exception is raised when no host key is found'''
    cnopts = CnOpts(knownhosts='sftpserver.pub')
    with pytest.raises(SSHException):
        cnopts.get_hostkey(host='missing-server')


def test_hostkey_returns_pkey():
    '''test the finding a matching host key returns a PKey'''
    cnopts = CnOpts(knownhosts='sftpserver.pub')
    assert isinstance(cnopts.get_hostkey('127.0.0.1'), RSAKey)



'''test sftpretty.remotetree'''

from common import conn, VFS
from pathlib import Path
from sftpretty import Connection
from tempfile import mkdtemp


def test_remotetree(sftpserver):
    '''test the remotetree function, with recurse'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            cwd = sftp.pwd
            localpath = Path(mkdtemp()).as_posix()
            tree = {}

            sftp.remotetree(tree, cwd, localpath)

            remote = {
                '/home/test': [
                    ('/home/test/pub', f'{localpath}/pub')
                ],
                '/home/test/pub': [
                    ('/home/test/pub/foo1',
                     f'{localpath}/pub/foo1'),
                    ('/home/test/pub/foo2',
                     f'{localpath}/pub/foo2')
                ],
                '/home/test/pub/foo2': [
                    ('/home/test/pub/foo2/bar1',
                     f'{localpath}/pub/foo2/bar1')
                ]
            }

            for branch in sorted(tree.keys()):
                assert set(remote[branch]) == set(tree[branch])
                del tree[branch]
            assert tree == {}


def test_remotetree_no_recurse(sftpserver):
    '''test the remotetree function, without recursing'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            cwd = sftp.pwd
            localpath = Path(mkdtemp()).as_posix()
            tree = {}

            sftp.remotetree(tree, cwd, localpath, recurse=False)

            remote = {
                '/home/test': [
                    ('/home/test/pub', f'{localpath}/pub')
                ]
            }

            for branch in sorted(tree.keys()):
                assert set(remote[branch]) == set(tree[branch])
                del tree[branch]
            assert tree == {}



'''test username parameter'''

# pylint: disable=W0212

import pytest

from common import conn, VFS
from os import environ
from sftpretty import Connection, CredentialException


def test_username_specified(sftpserver):
    '''test specifying username as parameter'''
    with sftpserver.serve_content(VFS):
        params = conn(sftpserver)
        params['username'] = 'bob'
        with Connection(**params) as sftp:
            assert sftp._username == params['username']


def test_username_from_environ(sftpserver):
    '''test reading username from $LOGNAME environment variable.'''
    username = 'bob'
    hold_logname = environ.get('LOGNAME')
    environ['LOGNAME'] = username
    with sftpserver.serve_content(VFS):
        params = conn(sftpserver)
        del params['username']
        with Connection(**params) as sftp:
            if hold_logname is not None:
                environ['LOGNAME'] = hold_logname
            assert sftp._username == username


def test_no_username_raises_err(sftpserver):
    '''test No username raises CredentialException.'''
    hold_logname = environ.get('LOGNAME')
    del environ['LOGNAME']
    with sftpserver.serve_content(VFS):
        params = conn(sftpserver)
        del params['username']
        with pytest.raises(CredentialException):
            sftp = Connection(**params)
            sftp.close()
    if hold_logname is not None:
        environ['LOGNAME'] = hold_logname



'''test sftpretty.execute'''


# TODO
# def test_execute_simple_ro(psftp):
#     '''test execute simple on a read-only server '''
#     results = [b'This service allows sftp connections only.\n', ]
#     assert psftp.execute('ls') == results


def test_execute_simple(lsftp):
    '''test execute simple'''
    type_check = type(b'')

    results = lsftp.execute('ls')
    # confirm results are an iterable of strings (version dependent)
    for result in results:
        assert isinstance(result, type_check)



'''test sftpretty.compression param'''

from common import LOCAL, SKIP_IF_CI
from sftpretty import CnOpts, Connection


def test_compression_default():
    '''test that a default connection does not have compression enabled'''
    with Connection(**LOCAL) as sftp:
        assert sftp.active_compression == ('none', 'none')


@SKIP_IF_CI
def test_compression_enabled():
    '''test that compress=True results in compression enabled, assuming
    that the server supports compression'''
    copts = LOCAL.copy()
    cnopts = CnOpts(knownhosts=None)
    cnopts.compress = True
    copts['cnopts'] = cnopts
    with Connection(**copts) as sftp:
        lcompress, rcompress = sftp.active_compression
        assert lcompress != 'none'
        assert rcompress != 'none'



'''test sftpretty.get'''

import pytest

from common import conn, tempfile_containing, VFS
from pathlib import Path
from sftpretty import Connection
from unittest.mock import Mock


def test_get(sftpserver):
    '''download a file'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo1')
            with tempfile_containing(contents='') as fname:
                sftp.get('foo1.txt', fname)
                assert open(fname, 'rb').read() == b'content of foo1.txt'


def test_get_bad_remote(sftpserver):
    '''download a file but it does not exist'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo1')
            with tempfile_containing(contents='') as fname:
                with pytest.raises(IOError):
                    sftp.get('readme-not-there.txt', fname)
                assert open(fname, 'rb').read()[0:7] != b'Welcome'


def test_get_callback(sftpserver):
    '''test .get callback'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo1')
            cback = Mock(return_value=None)
            with tempfile_containing(contents='') as fname:
                result = sftp.get('foo1.txt', fname, callback=cback)
                assert open(fname, 'rb').read() == b'content of foo1.txt'
            # verify callback was called
            assert cback.call_count
            # unlike .put() nothing is returned from the operation
            assert result is None


def test_get_glob_fails(sftpserver):
    '''try and use get a file with a pattern - Fails'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo1')
            with tempfile_containing(contents='') as fname:
                with pytest.raises(IOError):
                    sftp.get('*', fname)


def test_get_preserve_mtime(sftpserver):
    '''test that m_time is preserved from local to remote, when get'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo1')
            remotefile = 'foo1.txt'
            r_stat = sftp.stat(remotefile)
            with tempfile_containing(contents='') as localfile:
                sftp.get(remotefile, localfile, preserve_mtime=True)
                assert r_stat.st_mtime == Path(localfile).stat().st_mtime


def test_get_resume(sftpserver):
    '''test that resume continues partial download when it exists'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo1')
            remotesize = sftp.stat('foo1.txt').st_size
            with tempfile_containing(contents='content of') as fname:
                localsize = Path(fname).stat().st_size
                sftp.get('foo1.txt', fname, resume=True)
                assert open(fname, 'rb').read() == b'content of foo1.txt'
            # verify difference between remotesize and partial localsize
            assert 9 == (remotesize - localsize)



'''test sftpretty.remove and sftpretty.unlink methods'''

import pytest

from common import tempfile_containing
from pathlib import Path


def test_remove(lsftp):
    '''test the remove method'''
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        lsftp.chdir(Path.home().as_posix())
        lsftp.put(fname)
        is_there = base_fname in lsftp.listdir()
        lsftp.remove(base_fname)
        not_there = base_fname not in lsftp.listdir()

    assert is_there
    assert not_there


# TODO
# def test_remove_roserver(psftp):
#     '''test reaction of attempting remove on read-only server'''
#     psftp.chdir(Path.home().as_posix())
#     with pytest.raises(IOError):
#         psftp.remove('readme.txt')


def test_remove_does_not_exist(lsftp):
    '''test remove against a non-existant file'''
    lsftp.chdir(Path.home().as_posix())
    with pytest.raises(IOError):
        lsftp.remove('i-am-not-here.txt')



'''test sftpretty.normalize'''

from common import VFS, conn
from pathlib import Path
from sftpretty import Connection


def test_normalize(sftpserver):
    '''test the normalize function'''
    pubpath = Path('/home/test').joinpath('pub')
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            makepath = pubpath.parent.joinpath('make.txt').as_posix()
            assert sftp.normalize('make.txt') == makepath
            assert sftp.normalize('.') == pubpath.parent.as_posix()
            assert sftp.normalize('pub') == pubpath.as_posix()
            sftp.chdir('pub')
            assert sftp.normalize('.') == pubpath.as_posix()


# TODO
# def test_normalize_symlink(sftp):
#     '''test normalize against a symlink'''
#     home = Path.home()
#     sftp.chdir(home.as_posix())
#     rsym = 'readme.sym'
#     assert sftp.normalize(rsym) == home.joinpath(rsym).as_posix()


def test_pwd(sftpserver):
    '''test the pwd property'''
    pubpath = Path('/home/test').joinpath('pub')
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            sftp.chdir('pub/foo2')
            assert sftp.pwd == pubpath.joinpath('foo2').as_posix()
            sftp.chdir('bar1')
            assert sftp.pwd == pubpath.joinpath('foo2/bar1').as_posix()
            sftp.chdir('../../foo1')
            assert sftp.pwd == pubpath.joinpath('foo1').as_posix()



'''test sftpretty.getfo'''

from common import conn, VFS
from io import BytesIO
from sftpretty import Connection
from unittest.mock import Mock


def test_getfo_flo(sftpserver):
    '''test getfo to a file-like object'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            flo = BytesIO()
            sftp.chdir('pub')
            num_bytes = sftp.getfo('make.txt', flo)

            assert flo.getvalue() == b'content of make.txt'
            assert num_bytes == len(flo.getvalue())


def test_getfo_callback(sftpserver):
    '''test getfo callback'''
    with sftpserver.serve_content(VFS):
        with Connection(**conn(sftpserver)) as sftp:
            flo = BytesIO()
            cback = Mock(return_value=None)
            sftp.chdir('pub')
            sftp.getfo('make.txt', flo, callback=cback)

            assert cback.call_count



from concurrent.futures import as_completed, ThreadPoolExecutor
from contextlib import contextmanager
from functools import partial
from logging import (DEBUG, ERROR, FileHandler, Formatter, getLogger, INFO,
                     StreamHandler)
from os import environ, SEEK_END, utime
from paramiko import (Agent, hostkeys, SFTPClient, SSHConfig, Transport,
                      ConfigParseError, PasswordRequiredException,
                      SSHException, DSSKey, ECDSAKey, Ed25519Key, RSAKey)
from pathlib import Path
from sftpretty.exceptions import (CredentialException, ConnectionException,
                                  HostKeysException, LoggingException)
from sftpretty.helpers import _callback, drivedrop, hash, localtree, retry
from socket import gaierror
from stat import S_ISDIR, S_ISREG
from tempfile import mkstemp
from uuid import uuid4


class CnOpts(object):
    '''Additional connection options beyond authentication.

    :ivar tuple ciphers: *Default: paramiko.Transport.SecurityOptions.ciphers*
         - Ordered list of preferred ciphers for connection.
    :ivar bool compress: *Default: paramiko.Transport.use_compression* -
        Enable or disable compression.
    :ivar tuple compression:
         *Default: paramiko.Transport.SecurityOptions.compression* -
         Ordered tuple of preferred compression algorithms for connection.
    :ivar paramiko.SSHConfig: SSHConfig object used for parsing and
         preforming host based lookups on OpenSSH-style config directives.
    :ivar tuple digests: *Default: paramiko.Transport.SecurityOptions.digests*
         - Ordered tuple of preferred digests/macs for connection.
    :ivar dict disabled_algorithms: *Default: {}* - Mapping type to an
        iterable of algorithm identifiers, which will be disabled for the
        lifetime of the transport. Keys should match class builtin attribute.
    :ivar paramiko.hostkeys.HostKeys hostkeys: HostKeys object used for
        host key verifcation.
    :ivar tuple kex: *Default: paramiko.Transport.SecurityOptions.kex* -
        Ordered tuple of preferred key exchange algorithms for connection.
    :ivar tuple key_types:
         *Default: paramiko.Transport.SecurityOptions.key_types* -
         Ordered tuple of preferred public key types for connection.
    :param str config: *Default: ~/.ssh/config* - File path to load
        config from.
    :param str knownhosts: *Default: ~/.ssh/known_hosts* - File path to load
        hostkeys from.
    :ivar bool|str log: *Default: False* - Log connection details. If set to
        True, creates a temporary file used to capture logs. If set to an
        existing filepath, logs will be appended.
    :ivar str log_level: *Default: info* - Set logging level for connection.
        Choose between debug, error, or info.

    :returns: (obj) CnOpts - Connection options object, used for passing
        extended options to a Connection object.

    :raises ConfigParseError:
    :raises HostKeysException:
    '''
    def __init__(self, config=None, knownhosts=Path(
                 '~/.ssh/known_hosts').expanduser().as_posix()):
        self.ciphers = ('aes256-gcm@openssh.com', 'aes128-gcm@openssh.com',
                        'aes256-ctr', 'aes192-ctr', 'aes128-ctr', 'aes256-cbc',
                        'aes192-cbc', 'aes128-cbc', '3des-cbc')
        self.compress = False
        self.compression = ('none',)
        self.digests = ('hmac-sha2-512', 'hmac-sha2-256',
                        'hmac-sha2-512-etm@openssh.com',
                        'hmac-sha2-256-etm@openssh.com',
                        'hmac-sha1', 'hmac-md5')
        self.disabled_algorithms = {}
        self.hostkeys = hostkeys.HostKeys()
        self.kex = ('ecdh-sha2-nistp521', 'ecdh-sha2-nistp384',
                    'ecdh-sha2-nistp256', 'diffie-hellman-group16-sha512',
                    'diffie-hellman-group-exchange-sha256',
                    'diffie-hellman-group-exchange-sha1')
        self.key_types = ('ssh-ed25519', 'ecdsa-sha2-nistp521',
                          'ecdsa-sha2-nistp384', 'ecdsa-sha2-nistp256',
                          'rsa-sha2-512', 'rsa-sha2-256', 'ssh-rsa', 'ssh-dss')
        self.log = False
        self.log_level = 'info'
        self.ssh_config = SSHConfig()

        if config is not None:
            _config = Path(config).expanduser().resolve()
            if Path(_config).exists():
                self.ssh_config = self.ssh_config.from_path(_config)
            else:
                try:
                    self.ssh_config = self.ssh_config.from_file(_config)
                except ConfigParseError:
                    self.ssh_config = self.ssh_config.from_text(_config)
        else:
            _config = Path('~/.ssh/config').expanduser().resolve()
            if _config.exists():
                self.ssh_config = self.ssh_config.from_path(_config.as_posix())
            else:
                self.ssh_config = self.ssh_config.from_text('Host *')

        if knownhosts is not None:
            try:
                self.hostkeys.load(Path(knownhosts).resolve().as_posix())
            except FileNotFoundError:
                # no known_hosts in the default unix location
                raise UserWarning(
                    f'No file or host key found in [{knownhosts}]. '
                    'You will need to explicitly load host keys '
                    '(cnopts.hostkeys.load(filename)) or disable host '
                    'key verification (CnOpts(knownhosts=None)).'
                )
            else:
                if len(self.hostkeys.items()) == 0:
                    raise HostKeysException('No host keys found!')
        else:
            self.hostkeys = None

    def get_agentkey(self):
        '''Return the list of keys, if any, available through the local
        SSH agent. If no agent is running or one cannot be contacted, an
        empty tuple will be returned.

        :returns: (tuple of AgentKey objects) or (empty tuple)

        :raises SSHException:
        '''
        agent = Agent()
        keys = agent.get_keys()

        return keys

    def get_config(self, host):
        '''Return config options for a given host-match.

        :param str host: Identifier to lookup using OpenSSH's ssh_config
            man page ruleset. The first value matched will be returned.

        :returns: (obj) SSHConfigDict - A dictionary wrapper/subclass for
            per-host configuration structures.
        '''
        cval = self.ssh_config.lookup(host)
        return cval or {}

    def get_hostkey(self, host):
        '''Return the matching known hostkey to be used for verification or
        raise an SSHException.

        :param str host: The Hostname or IP of the remote machine.

        :returns: (obj) PKey - Public key(s) associated with host or None.

        :raises SSHException:
        '''
        kval = self.hostkeys.lookup(host)
        # None | {key_type: private_key}
        if kval is None:
            raise SSHException(f'No hostkey for host [{host}] found.')

        # Return the public key from the dictionary
        return list(kval.values())[0]


class Connection(object):
    '''Connects and logs into the specified hostname. Arguments that are not
    given are guessed from the environment.

    :param str host: *Required* - Hostname or address of the remote machine.
    :param CnOpts|None cnopts: *Default: None* - Extended connection options
        set as a CnOpts object.
    :param str|None default_path: *Default: None* - Set the default working
        directory upon connection.
    :param str|None password: *Default: None* - Credential for remote machine.
    :param int port: *Default: 22* - SFTP server port of the remote machine.
    :param str|obj|None private_key: *Default: None* - Path to private key
        file(str) or paramiko.AgentKey object
    :param str|None private_key_pass: *Default: None* - Password to use on
        encrypted private_key.
    :param float|None timeout: *Default: None* - Set channel timeout.
    :param str|None username: *Default: None* - User for remote machine.

    :returns: (obj) Connection to the requested host.

    :raises ConnectionException:
    :raises CredentialException:
    :raises HostKeysException:
    :raises LoggingException:
    :raises PasswordRequiredException:
    :raises SSHException:
    '''
    def __init__(self, host, cnopts=None, default_path=None, password=None,
                 port=22, private_key=None, private_key_pass=None,
                 timeout=None, username=None):
        self._cnopts = cnopts or CnOpts()
        self._config = self._cnopts.get_config(host)
        self._default_path = default_path
        self._set_logging()
        self._timeout = self._config.get('connecttimeout') or timeout
        self._transport = None
        self._start_transport(self._config.get('hostname') or host,
                              self._config.get('port') or port)
        self._set_username(self._config.get('user') or username)
        self._set_authentication(password, private_key, private_key_pass)

    def _set_authentication(self, password, private_key, private_key_pass):
        '''Authenticate transport. Prefer private key over password.'''
        if self._config.get('identityfile'):
            private_key = self._config['identityfile'][0]
        if private_key is not None:
            # Use key path or provided key object
            key_types = {'DSA': DSSKey, 'EC': ECDSAKey, 'OPENSSH': Ed25519Key,
                         'RSA': RSAKey}
            if isinstance(private_key, str):
                key_file = Path(private_key).expanduser().absolute().as_posix()
                try:
                    with open(key_file, 'r', encoding='utf-8') as head:
                        key_id = head.readline()[11:][:-18]
                    log.debug(f'Key ID: [{key_id}]')
                    key = key_types[key_id.strip()]
                except KeyError as err:
                    log.error(('Unable to identify key type from file provided'
                              f': \n[{key_file}]'))
                    raise err
                except PasswordRequiredException as err:
                    log.error(('No password provided for encrypted private '
                               'key encrypted private key.'))
                    raise err
                except PermissionError as err:
                    log.error(('File permission preventing user access to:\n'
                              f'[{key_file}]'))
                    raise err
                except SSHException as err:
                    log.error(('Path provided is an invalid key file, a '
                               'directory or does not exist, please revise '
                               'and provide a path to a valid private key.'))
                    raise err
                finally:
                    private_key = key.from_private_key_file(
                        key_file, password=private_key_pass)
            self._transport.auth_publickey(self._username, private_key)
        elif password is not None:
            self._transport.auth_password(self._username, password)
        else:
            raise CredentialException('No password or private key provided.')

    def _set_logging(self):
        '''Set logging location and level for connection'''
        level_map = {'debug': DEBUG, 'error': ERROR, 'info': INFO}
        level = self._config.get('loglevel') or self._cnopts.log_level
        level = level_map[level.lower().strip('1,2,3')]

        try:
            global log
            log = getLogger('SFTPretty')
            if self._cnopts.log:
                if isinstance(self._cnopts.log, bool):
                    # Log to a temporary file.
                    flo, self._cnopts.log = mkstemp('.txt', 'sftpretty-')
                logfile = FileHandler(self._cnopts.log, encoding='utf8')
                logfile.setLevel = (level)
                logfile_formatter = Formatter(('[%(asctime)s] %(levelname)s - '
                                               '%(message)s'))
                logfile.setFormatter(logfile_formatter)
                log.addHandler(logfile)
            console = StreamHandler()
            console.setLevel(level)
            console_formatter = Formatter(('[%(asctime)s] %(levelname)s - '
                                           '%(message)s'))
            console.setFormatter(console_formatter)
            log.addHandler(console)
            log.setLevel(level)
        except KeyError:
            raise LoggingException(('Log level must set to one of following: '
                                    '[debug, error, info].'))

    def _set_username(self, username):
        '''Set the username for the connection. If not passed, then look to
        the environment. Still nothing? Throw CredentialException.'''
        local_username = environ.get('LOGNAME', None)

        if username is not None:
            self._username = username
        elif local_username is not None:
            self._username = local_username
        else:
            raise CredentialException('No username specified.')

    @contextmanager
    def _sftp_channel(self, keepalive=False):
        '''Establish new SFTP channel.'''
        _channel = None

        try:
            _channel = SFTPClient.from_transport(self._transport)

            channel = _channel.get_channel()
            channel_name = uuid4().hex
            channel.set_name(channel_name)
            channel.settimeout(self._timeout)
            log.debug(f'Channel Name: [{channel_name}]')

            if self._default_path is not None:
                _channel.chdir(drivedrop(self._default_path))
                log.info(f'Current Working Directory: [{self._default_path}]')

            yield _channel
        except Exception as err:
            raise err
        finally:
            if _channel and not keepalive:
                _channel.close()

    def _start_transport(self, host, port):
        '''Start the transport and set connection options if specified.'''
        try:
            self._transport = Transport((host, int(port)))

            keepalive = self._config.get('serveraliveinterval') or 60
            self._transport.set_keepalive(int(keepalive))
            self._transport.set_log_channel(host)

            compress = self._config.get('compression') or self._cnopts.compress
            self._transport.use_compression(compress=bool(compress))

            # Set disabled algorithms
            disabled_algorithms = self._cnopts.disabled_algorithms
            self._transport.disabled_algorithms = disabled_algorithms
            log.debug(f'Disabled Algorithms: [{disabled_algorithms}]')

            # Security Options
            # Set allowed ciphers
            ciphers = self._config.get('ciphers') or self._cnopts.ciphers
            _ciphers = self._transport.get_security_options().ciphers
            if not isinstance(ciphers, tuple):
                ciphers = tuple(ciphers.split(','))
            self._transport.get_security_options().ciphers = tuple(
                cipher for cipher in ciphers if cipher in _ciphers)
            log.debug(f'Ciphers: [{ciphers}]')
            # Set compression algorithms
            compression = self._cnopts.compression
            self._transport.get_security_options().compression = compression
            log.debug(f'Compression: [{compression}]')
            # Set connection digests
            digests = self._config.get('macs') or self._cnopts.digests
            _digests = self._transport.get_security_options().digests
            if not isinstance(digests, tuple):
                digests = tuple(digests.split(','))
            self._transport.get_security_options().digests = tuple(
                digest for digest in digests if digest in _digests)
            log.debug(f'MACs: [{digests}]')
            # Set connection kex
            kexs = self._config.get('kexalgorithms') or self._cnopts.kex
            _kex = self._transport.get_security_options().kex
            if not isinstance(kexs, tuple):
                kexs = tuple(kexs.split(','))
            self._transport.get_security_options().kex = tuple(
                kex for kex in kexs if kex in _kex)
            log.debug(f'KEX: [{kexs}]')
            # Set allowed key types
            key_types = self._config.get('pubkeyacceptedalgorithms') or\
                self._cnopts.key_types
            _key_types = self._transport.get_security_options().key_types
            if not isinstance(key_types, tuple):
                key_types = tuple(key_types.split(','))
            self._transport.get_security_options().key_types = tuple(
                key_type for key_type in key_types if key_type in _key_types)
            log.debug(f'Public Key Types: [{key_types}]')

            self._transport.start_client(timeout=self._timeout)

            if self._transport.is_active():
                remote_hostkey = self._transport.get_remote_server_key()
                remote_fingerprint = hash(remote_hostkey)
                log.info((f'[{host}] Host Key: \n\t'
                          f'Name: {remote_hostkey.get_name()}\n\t'
                          f'Fingerprint: {remote_fingerprint}\n\t'
                          f'Size: {remote_hostkey.get_bits():d}'))

                if self._cnopts.hostkeys is not None:
                    user_hostkey = self._cnopts.get_hostkey(host)
                    user_fingerprint = hash(user_hostkey)
                    log.info(f'Known Fingerprint: {user_fingerprint}')
                    if user_fingerprint != remote_fingerprint:
                        raise HostKeysException((f'{host} key verification: '
                                                 '[FAILED]'))
            else:
                err = self._transport.get_exception()
                if err:
                    self.close()
                    raise err
        except (AttributeError, gaierror, UnicodeError):
            raise ConnectionException(host, port)
        except Exception as err:
            raise err

    def get(self, remotefile, localpath=None, callback=None,
            max_concurrent_prefetch_requests=None, prefetch=True,
            preserve_mtime=False, resume=False, exceptions=None, tries=None,
            backoff=2, delay=1, logger=getLogger(__name__), silent=False):
        '''Copies a file between the remote host and the local host.

        :param str remotefile: The remote path and filename to retrieve.
        :param str localpath: The local path to save download.
            If None, file is copied to local current working directory.
        :param callable callback: Optional callback function (form: ``func(
            int, int)``) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param bool preserve_mtime: *Default: False* - Sync the modification
            time(st_mtime) on the local file to match the time on the remote.
            (st_atime can differ because stat'ing the localfile can/does update
            it's st_atime).
        :param int max_concurrent_prefetch_requests: *Default: None* - The
            maximum number of concurrent read requests to prefetch.
        :param bool prefetch: *Default: True* - Controls whether prefetching
            is performed.
        :param bool resume: *Default: False* - Continue a previous transfer
            based on destination path matching.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: None

        :raises: IOError
        '''
        @retry(exceptions, tries=tries, backoff=backoff, delay=delay,
               logger=logger, silent=silent)
        def _get(self, remotefile, localpath=None, callback=None,
                 max_concurrent_prefetch_requests=None, prefetch=True,
                 preserve_mtime=False, resume=False):

            if localpath is None:
                localpath = Path(remotefile).name

            if callback is None:
                callback = partial(_callback, remotefile, logger=logger)

            with self._sftp_channel() as channel:
                if resume:
                    if Path(localpath).is_file():
                        localsize = Path(localpath).stat().st_size
                        log.info((f'Resuming existing download of {localpath} '
                                  f'@ {localsize} bytes'))
                    else:
                        localsize = 0
                    remote_attributes = remotesize = channel.stat(remotefile)
                    log.debug(f'[{remotesize.st_size}]: {remotefile}')
                    if localsize < remotesize.st_size:
                        with open(localpath, 'ab') as localfile:
                            with channel.open(remotefile, 'rb') as remotepath:
                                if localsize > 0:
                                    remotepath.seek(localsize)
                                if prefetch:
                                    remotepath.prefetch(remotesize.st_size,
                                                        max_concurrent_prefetch_requests)  # noqa: E501
                                channel._transfer_with_callback(
                                    callback=callback,
                                    file_size=remotesize.st_size,
                                    reader=remotepath,
                                    writer=localfile)
                else:
                    if preserve_mtime:
                        remote_attributes = channel.stat(remotefile)

                    channel.get(remotefile, localpath=localpath,
                                callback=callback, prefetch=prefetch,
                                max_concurrent_prefetch_requests=max_concurrent_prefetch_requests)  # noqa: E501

            if preserve_mtime:
                utime(localpath, (remote_attributes.st_atime,
                                  remote_attributes.st_mtime))

        _get(self, remotefile, localpath=localpath, callback=callback,
             max_concurrent_prefetch_requests=max_concurrent_prefetch_requests,
             prefetch=prefetch, preserve_mtime=preserve_mtime, resume=resume)

    def get_d(self, remotedir, localdir, callback=None,
              max_concurrent_prefetch_requests=None, pattern=None,
              prefetch=True, preserve_mtime=False, resume=False, workers=None,
              exceptions=None, tries=None, backoff=2, delay=1,
              logger=getLogger(__name__), silent=False):
        '''Get the contents of remotedir and write to locadir. Non-recursive.

        :param str remotedir: The remote directory to copy locally.
        :param str localdir: The local path to save download.
        :param callable callback: Optional callback function (form: ``func(
            int, int``)) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param int max_concurrent_prefetch_requests: *Default: None* - The
            maximum number of concurrent read requests to prefetch.
        :param str pattern: *Default: None* - Filter applied to filenames to
            transfer only subset of files in a directory.
        :param bool prefetch: *Default: True* - Controls whether prefetching
            is performed.
        :param bool preserve_mtime: *Default: False* - Sync the modification
            time(st_mtime) on the local file to match the time on the remote.
            (st_atime can differ because stat'ing the localfile can/does update
            it's st_atime)
        :param bool resume: *Default: False* - Continue a previous transfer
            based on destination path matching.
        :param int workers: *Default: None* - If None, defaults to number of
            processors plus 4. Set to less than or equal to allowed
            concurrent connections on server.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: None

        :raises: Any exception raised by operations will be passed through.
        '''
        filelist = self.listdir_attr(remotedir)

        if not Path(localdir).is_dir():
            Path(localdir).mkdir(exist_ok=True, parents=True)
            logger.info(f'Creating Folder [{localdir}]!')

        if pattern is None:
            paths = [
                (Path(remotedir).joinpath(attribute.filename).as_posix(),
                 Path(localdir).joinpath(attribute.filename).as_posix(),
                 callback, max_concurrent_prefetch_requests, prefetch,
                 preserve_mtime, resume, exceptions, tries, backoff,
                 delay, logger, silent)
                for attribute in filelist if S_ISREG(attribute.st_mode)
            ]
        else:
            paths = [
                (Path(remotedir).joinpath(attribute.filename).as_posix(),
                 Path(localdir).joinpath(attribute.filename).as_posix(),
                 callback, max_concurrent_prefetch_requests, prefetch,
                 preserve_mtime, resume, exceptions, tries, backoff,
                 delay, logger, silent)
                for attribute in filelist if S_ISREG(attribute.st_mode)
                if f'{pattern}' in attribute.filename
            ]

        if paths != []:
            thread_prefix = uuid4().hex
            with ThreadPoolExecutor(max_workers=workers,
                                    thread_name_prefix=thread_prefix) as pool:
                logger.debug(f'Thread Prefix: [{thread_prefix}]')
                threads = {
                           pool.submit(self.get, remote, local,
                                       callback=callback,
                                       max_concurrent_prefetch_requests=max_concurrent_prefetch_requests,  # noqa: E501
                                       prefetch=prefetch, resume=resume,
                                       preserve_mtime=preserve_mtime,
                                       exceptions=exceptions, tries=tries,
                                       backoff=backoff, delay=delay,
                                       logger=logger, silent=silent): remote
                           for remote, local, callback,
                           max_concurrent_prefetch_requests, prefetch,
                           preserve_mtime, resume, exceptions, tries, backoff,
                           delay, logger, silent in paths
                          }
                for future in as_completed(threads):
                    name = threads[future]
                    try:
                        data = future.result()
                    except Exception as err:
                        logger.error(f'Thread [{name}]: [FAILED]')
                        raise err
                    else:
                        logger.info(f'Thread [{name}]: [COMPLETE]')
                        return data
        else:
            logger.info(f'No files found in directory [{remotedir}]')

    def get_r(self, remotedir, localdir, callback=None,
              max_concurrent_prefetch_requests=None, pattern=None,
              prefetch=True, preserve_mtime=False, resume=False, workers=None,
              exceptions=None, tries=None, backoff=2, delay=1,
              logger=getLogger(__name__), silent=False):
        '''Recursively copy remotedir structure to localdir

        :param str remotedir: The remote directory to recursively copy.
        :param str localdir: The local path to save recursive download.
        :param callable callback: Optional callback function (form: ``func(
            int, int``)) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param int max_concurrent_prefetch_requests: *Default: None* - The
            maximum number of concurrent read requests to prefetch.
        :param str pattern: *Default: None* - Filter applied to all filenames
            transfering only the subset of files that match.
        :param bool prefetch: *Default: True* - Controls whether prefetching
            is performed.
        :param bool preserve_mtime: *Default: False* - Sync the modification
            time(st_mtime) on the local file to match the time on the remote.
            (st_atime can differ because stat'ing the localfile can/does update
            it's st_atime)
        :param bool resume: *Default: False* - Continue a previous transfer
            based on destination path matching.
        :param int workers: *Default: None* - If None, defaults to number of
            processors plus 4. Set to less than or equal to allowed
            concurrent connections on server.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: None

        :raises: Any exception raised by operations will be passed through.
        '''
        self.chdir(remotedir)

        lwd = Path(localdir).absolute().as_posix()
        rwd = self._default_path

        tree = {}
        tree[rwd] = [(rwd, lwd)]

        self.remotetree(tree, rwd, lwd, recurse=True)
        log.debug(f'Remote Tree: [{tree}]')

        for roots in tree.keys():
            for remote, local in tree[roots]:
                self.get_d(remote, local, callback=callback,
                           max_concurrent_prefetch_requests=max_concurrent_prefetch_requests,  # noqa: E501
                           pattern=pattern, prefetch=prefetch,
                           preserve_mtime=preserve_mtime, resume=resume,
                           workers=workers, exceptions=exceptions, tries=tries,
                           backoff=backoff, delay=delay, logger=logger,
                           silent=silent)

    def getfo(self, remotefile, flo, callback=None,
              max_concurrent_prefetch_requests=None, prefetch=True,
              exceptions=None, tries=None, backoff=2, delay=1,
              logger=getLogger(__name__), silent=False):
        '''Copy a remote file (remotepath) to a file-like object, flo.

        :param str remotefile: The remote path and filename to retrieve.
        :param flo: Open file like object ready to write.
        :param callable callback: Optional callback function (form: ``func(
            int, int``)) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param int max_concurrent_prefetch_requests: *Default: None* - The
            maximum number of concurrent read requests to prefetch.
        :param bool prefetch: *Default: True* - Controls whether prefetching
            is performed.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: (int) The number of bytes written to the opened file object

        :raises: Any exception raised by operations will be passed through.
        '''
        @retry(exceptions, tries=tries, backoff=backoff, delay=delay,
               logger=logger, silent=silent)
        def _getfo(self, remotefile, flo, callback=None,
                   max_concurrent_prefetch_requests=None, prefetch=True):

            if callback is None:
                callback = partial(_callback, remotefile, logger=logger)

            with self._sftp_channel() as channel:
                flo_size = channel.getfo(remotefile, flo, callback=callback,
                                         max_concurrent_prefetch_requests=max_concurrent_prefetch_requests,  # noqa: E501
                                         prefetch=prefetch)

            return flo_size

        return _getfo(self, remotefile, flo, callback=callback,
                      max_concurrent_prefetch_requests=max_concurrent_prefetch_requests,  # noqa: E501
                      prefetch=prefetch)

    def put(self, localfile, remotepath=None, callback=None, confirm=True,
            preserve_mtime=False, resume=False, exceptions=None, tries=None,
            backoff=2, delay=1, logger=getLogger(__name__), silent=False):
        '''Copies a file between the local host and the remote host.

        :param str localfile: The local path and filename to copy remotely.
        :param str remotepath: Remote location to save file, else the remote
            :attr:`.pwd` and local filename is used.
        :param callable callback: Optional callback function (form: ``func(
            int, int``)) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param bool confirm: *Default: True* - Whether to do a stat() on the
            file afterwards to the file size.
        :param bool preserve_mtime: *Default: False* - Make the modification
            time(st_mtime) on the remote file match the time on the local.
            (st_atime can differ because stat'ing the localfile can/does update
            it's st_atime)
        :param bool resume: *Default: False* - Continue a previous transfer
            based on destination path matching.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: (obj) SFTPAttributes containing details about the given file.

        :raises IOError: if remotepath doesn't exist
        :raises OSError: if localfile doesn't exist
        '''
        @retry(exceptions, tries=tries, backoff=backoff, delay=delay,
               logger=logger, silent=silent)
        def _put(self, localfile, remotepath=None, callback=None,
                 confirm=True, preserve_mtime=False, resume=False):

            if remotepath is None:
                remotepath = Path(localfile).name

            if callback is None:
                callback = partial(_callback, localfile, logger=logger)

            if preserve_mtime:
                local_attributes = Path(localfile).stat()
                local_times = (local_attributes.st_atime,
                               local_attributes.st_mtime)

            with self._sftp_channel() as channel:
                remotepath = drivedrop(remotepath)
                if resume:
                    remote = channel.stat(remotepath)
                    if S_ISREG(remote.st_mode):
                        remotesize = remote.st_size
                        log.info((f'Resuming existing upload of {remotepath} '
                                  f'@ {remotesize} bytes'))
                    else:
                        remotesize = 0
                    localsize = Path(localfile).stat().st_size
                    log.debug(f'[{localsize}]: {localfile}')
                    if localsize > remotesize:
                        with channel.open(remotepath, 'ab') as remotefile:
                            remotefile.set_pipelined(True)
                            with open(localfile, 'rb') as localpath:
                                if remotesize > 0:
                                    localpath.seek(remotesize)
                                resumesize = channel._transfer_with_callback(
                                    callback=callback, file_size=localsize,
                                    reader=localpath, writer=remotefile)
                    if confirm:
                        attributes = channel.stat(remotepath)
                        if attributes.st_size != (remotesize + resumesize):
                            raise IOError(('size mismatch in put! '
                                           f'{attributes.st_size} != '
                                           f'{remotesize}'))

                else:
                    attributes = channel.put(localfile, remotepath=remotepath,
                                             callback=callback,
                                             confirm=confirm)

                if preserve_mtime:
                    channel.utime(remotepath, local_times)
                    attributes = channel.stat(remotepath)

            return attributes

        return _put(self, localfile, remotepath=remotepath, callback=callback,
                    confirm=confirm, preserve_mtime=preserve_mtime,
                    resume=resume)

    def put_d(self, localdir, remotedir, callback=None, confirm=True,
              preserve_mtime=False, resume=False, workers=None,
              exceptions=None, tries=None, backoff=2, delay=1,
              logger=getLogger(__name__), silent=False):
        '''Copies a local directory's contents to a remotepath

        :param str localdir: The local directory to copy remotely.
        :param str remotedir: The remote location to save directory.
        :param callable callback: Optional callback function (form: ``func(
            int, int``)) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param bool confirm: *Default: True* - Whether to do a stat() on the
            file afterwards to confirm the file size.
        :param bool preserve_mtime: *Default: False* - Make the modification
            time(st_mtime) on the remote file match the time on the local.
            (st_atime can differ because stat'ing the localfile can/does update
            it's st_atime)
        :param bool resume: *Default: False* - Continue a previous transfer
            based on destination path matching.
        :param int workers: *Default: None* - If None, defaults to number of
            processors plus 4. Set to less than or equal to allowed
            concurrent connections on server.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: None

        :raises IOError: if remotedir doesn't exist
        :raises OSError: if localdir doesn't exist
        '''
        localdir = Path(localdir)

        self.mkdir_p(Path(remotedir).joinpath(localdir.stem).as_posix())

        paths = [
            (localpath.as_posix(),
             Path(remotedir).joinpath(localpath.relative_to(
                 localdir.parent).as_posix()).as_posix(),
             callback, confirm, preserve_mtime, resume, exceptions, tries,
             backoff, delay, logger, silent)
            for localpath in localdir.iterdir()
            if localpath.is_file()
        ]

        if paths != []:
            thread_prefix = uuid4().hex
            with ThreadPoolExecutor(max_workers=workers,
                                    thread_name_prefix=thread_prefix) as pool:
                logger.debug(f'Thread Prefix: [{thread_prefix}]')
                threads = {
                    pool.submit(
                        self.put, local, remote, callback=callback,
                        confirm=confirm, preserve_mtime=preserve_mtime,
                        resume=resume, exceptions=exceptions, tries=tries,
                        backoff=backoff, delay=delay, logger=logger,
                        silent=silent
                    ): local
                    for local, remote, callback, confirm,
                    preserve_mtime, resume, exceptions, tries, backoff,
                    delay, logger, silent in paths
                }

                for future in as_completed(threads):
                    name = threads[future]
                    try:
                        data = future.result()
                    except Exception as err:
                        logger.error(f'Thread [{name}]: [FAILED]')
                        raise err
                    else:
                        logger.info(f'Thread [{name}]: [COMPLETE]')
                        return data
        else:
            logger.info(f'No files found in directory [{localdir}]')

    def put_r(self, localdir, remotedir, callback=None, confirm=True,
              preserve_mtime=False, resume=False, workers=None,
              exceptions=None, tries=None, backoff=2, delay=1,
              logger=getLogger(__name__), silent=False):
        '''Recursively copies a local directory's contents to a remotepath

        :param str localdir: The local directory to copy remotely.
        :param str remotedir: The remote location to save directory.
        :param callable callback: Optional callback function (form: ``func(
            int, int``)) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param bool confirm: *Default: True* - Whether to do a stat() on the
            file afterwards to confirm the file size.
        :param bool preserve_mtime: *Default: False* - Make the modification
            time(st_mtime) on the remote file match the time on the local.
            (st_atime can differ because stat'ing the localfile can/does update
            it's st_atime)
        :param bool resume: *Default: False* - Continue a previous transfer
            based on destination path matching.
        :param int workers: *Default: None* - If None, defaults to number of
            processors plus 4. Set to less than or equal to allowed
            concurrent connections on server.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: None

        :raises IOError: if remotedir doesn't exist
        :raises OSError: if localdir doesn't exist
        '''
        lwd = Path(localdir).absolute().as_posix()
        rwd = self.normalize(remotedir)

        tree = {}
        tree[lwd] = [(lwd, rwd)]

        localtree(tree, lwd, rwd, recurse=True)
        log.debug(f'Local Tree: [{tree}]')

        for roots in tree.keys():
            for local, remote in tree[roots]:
                self.put_d(local, remote, callback=callback, confirm=confirm,
                           preserve_mtime=preserve_mtime, resume=resume,
                           workers=workers, exceptions=exceptions, tries=tries,
                           backoff=backoff, delay=delay, logger=logger,
                           silent=silent)

    def putfo(self, flo, remotepath=None, file_size=None, callback=None,
              confirm=True, exceptions=None, tries=None, backoff=2,
              delay=1, logger=getLogger(__name__), silent=False):
        '''Copies the contents of a file like object to remotepath.

        :param flo: File-like object that supports .read()
        :param str remotepath: The remote location to save contents of object.
        :param int file_size: The size of flo, if not given, calculated
            preventing division by zero in default callback function.
        :param callable callback: Optional callback function (form: ``func(
            int, int``)) that accepts the bytes transferred so far and the
            total bytes to be transferred.
        :param bool confirm: *Default: True* - Whether to do a stat() on the
            file afterwards to confirm the file size.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: Times to try (not retry) before giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: (obj) SFTPAttributes containing details about the given file.

        :raises: TypeError, if remotepath not specified, any underlying error
        '''
        @retry(exceptions, tries=tries, backoff=backoff, delay=delay,
               logger=logger, silent=silent)
        def _putfo(self, flo, remotepath=None, file_size=None, callback=None,
                   confirm=True):

            if callback is None:
                callback = partial(_callback, flo, logger=logger)

            if file_size is None:
                file_size = flo.seek(0, SEEK_END)
                flo.seek(0)

            if remotepath is None:
                remotepath = uuid4().hex

            with self._sftp_channel() as channel:
                attributes = channel.putfo(flo, remotepath=remotepath,
                                           file_size=file_size,
                                           callback=callback, confirm=confirm)

            return attributes

        return _putfo(self, flo, remotepath=remotepath, file_size=file_size,
                      callback=callback, confirm=confirm)

    def execute(self, command,
                exceptions=None, tries=None, backoff=2, delay=1,
                logger=getLogger(__name__), silent=False):
        '''Execute the given commands on a remote machine.  The command is
        executed without regard to the remote :attr:`.pwd`.

        :param str command: Command to execute.
        :param Exception exceptions: Exception(s) to check. May be a tuple of
            exceptions to check. IOError or IOError(errno.ECOMM) or (IOError,)
            or (ValueError, IOError(errno.ECOMM))
        :param int tries: *Default: None* - Times to try (not retry) before
            giving up.
        :param int backoff: *Default: 2* - Backoff multiplier. Default will
            double the delay each retry.
        :param int delay: *Default: 1* - Initial delay between retries in
            seconds.
        :param logging.Logger logger: *Default: Logger(__name__)* -
            Logger to use.
        :param bool silent: *Default: False* - If set then no logging will
            be attempted.

        :returns: (list of str) Results of the command.

        :raises: Any exception raised by command will be passed through.
        '''
        @retry(exceptions, backoff=backoff, delay=delay, logger=logger,
               silent=silent, tries=tries)
        def _execute(self, command):
            channel = self._transport.open_session()
            channel.exec_command(command)

            output = channel.makefile('rb', -1).readlines()

            if output:
                return output
            else:
                return channel.makefile_stderr('rb', -1).readlines()

        return _execute(self, command)

    @contextmanager
    def cd(self, remotepath=None):
        '''Context manager that can change to a optionally specified remote
        directory and restores the old pwd on exit.

        :param str|None remotepath: *Default: None* - Remote path to maintain
            as the current working directory.

        :returns: None

        :raises: IOError, if remote path doesn't exist
        '''
        original_path = self.pwd

        try:
            if remotepath is not None:
                self.chdir(remotepath)
            yield
        except Exception as err:
            raise err
        finally:
            self.chdir(original_path)

    def chdir(self, remotepath):
        '''Change the current working directory on the remote

        :param str remotepath: Remote path to set as current working directory.

        :returns: None

        :raises: IOError, if path does not exist
        '''
        with self._sftp_channel() as channel:
            channel.chdir(drivedrop(remotepath))
            self._default_path = channel.normalize('.')

    def chmod(self, remotepath, mode=700):
        '''Set the permission mode of a remotepath, where mode is an octal.

        :param str remotepath: Remote path to modify permission.
        :param int mode: *Default: 700* - Octal mode to apply on path.

        :returns: None

        :raises: IOError, if the file doesn't exist
        '''
        with self._sftp_channel() as channel:
            channel.chmod(drivedrop(remotepath), mode=int(str(mode), 8))

    def chown(self, remotepath, uid=None, gid=None):
        '''Set uid/gid on remotepath, you may specify either or both.

        :param str remotepath: Remote path to modify ownership.
        :param int uid: User id to set as owner of remote path.
        :param int gid: Group id to set on the remote path.

        :returns: None

        :raises: IOError, if user lacks permission or if the file doesn't exist
        '''
        with self._sftp_channel() as channel:
            remotepath = drivedrop(remotepath)
            if uid is None or gid is None:
                if uid is None and gid is None:
                    return
                remote_attributes = channel.stat(remotepath)
                if uid is None:
                    uid = remote_attributes.st_uid
                if gid is None:
                    gid = remote_attributes.st_gid

            channel.chown(remotepath, uid=uid, gid=gid)

    def close(self):
        '''Terminate transport connection and clean up the bits.'''
        try:
            # Close the transport.
            if self._transport and self._transport.is_active():
                self._transport.close()
            self._transport = None
            # Clean up any loggers
            if log.hasHandlers():
                # remove lingering handlers if any
                for handle in log.handlers:
                    log.removeHandler(handle)
        except AttributeError:
            pass
        except Exception as err:
            raise err

    def exists(self, remotepath):
        '''Test whether a remotepath exists.

        :param str remotepath: Remote location to verify existance of.

        :returns: (bool) True, if remotepath exists, else False.
        '''
        with self._sftp_channel() as channel:
            try:
                channel.stat(remotepath)
            except IOError as err:
                if err.errno == 2:
                    return False
                else:
                    raise err

            return True

    def getcwd(self):
        '''Return the current working directory on the remote.

        :returns: (str) Remote current working directory. None, if not set.
        '''
        with self._sftp_channel() as channel:
            cwd = channel.getcwd()

        return cwd

    def isdir(self, remotepath):
        '''Determine if remotepath is a directory.

        :param str remotepath: Remote location to test.

        :returns: (bool)
        '''
        with self._sftp_channel() as channel:
            try:
                result = S_ISDIR(channel.stat(remotepath).st_mode)
            except IOError:
                # No such directory
                result = False

        return result

    def isfile(self, remotepath):
        '''Determine if remotepath is a file.

        :param str remotepath: Remote location to test.

        :returns: (bool)
        '''
        with self._sftp_channel() as channel:
            try:
                result = S_ISREG(channel.stat(remotepath).st_mode)
            except IOError:
                # No such file
                result = False

        return result

    def lexists(self, remotepath):
        '''Determine whether remotepath exists.

        :param str remotepath: Remote location to test.

        :returns: (bool), True, if lexists, else False
        '''
        with self._sftp_channel() as channel:
            try:
                channel.lstat(drivedrop(remotepath))
            except IOError:
                return False

        return True

    def listdir(self, remotepath='.'):
        '''Return a sorted list of a directory's contents.

        :param str remotepath: Remote location to search.

        :returns: (list of str) Sorted directory content.

        '''
        with self._sftp_channel() as channel:
            directory = sorted(channel.listdir(drivedrop(remotepath)))

        return directory

    def listdir_attr(self, remotepath='.'):
        '''Return a non-sorted list of SFTPAttribute objects for the remote
        directory contents. Will not include the special entries '.' and '..'.

        The returned SFTPAttributes objects will each have an additional field:
        longname, which may contain a formatted string of the file's
        attributes, in unix format. The content of this string will depend on
        the SFTP server.

        :param str remotepath: Remote location to search.

        :returns: (list of SFTPAttributes) Sorted directory content as objects.
        '''
        with self._sftp_channel() as channel:
            directory = sorted(channel.listdir_attr(drivedrop(remotepath)),
                               key=lambda attribute: attribute.filename)

        return directory

    def lstat(self, remotepath):
        '''Return information about remote location without following symbolic
        links. Otherwise, the same as .stat().

        :param str remotepath: Remote location to stat.

        :returns: (obj) SFTPAttributes object
        '''
        with self._sftp_channel() as channel:
            lstat = channel.lstat(drivedrop(remotepath))

        return lstat

    def mkdir(self, remotedir, mode=700):
        '''Create a directory and set permission mode. On some systems, mode
        is ignored. Where used, the current umask value is first masked out.

        :param str remotedir: Remote location to create.
        :param int mode: *Default: 700* - Octal mode to apply on path.

        :returns: None
        '''
        with self._sftp_channel() as channel:
            channel.mkdir(drivedrop(remotedir), mode=int(str(mode), 8))

    def mkdir_p(self, remotedir, mode=700):
        '''Create a directory and any missing parent locations as needed. Set
        permission mode, if created. Silently complete if remotedir already
        exists.

        :param str remotedir: Remote location to create.
        :param int mode: *Default: 700* - Octal mode to apply on created paths.

        :returns: None

        :raises: OSError
        '''
        try:
            remotedir = drivedrop(remotedir)
            if self.isdir(remotedir):
                return
            elif self.isfile(remotedir):
                raise OSError((f'A file with the same name, [{remotedir}], '
                               'already exists.'))
            else:
                parent = Path(remotedir).parent.as_posix()
                stem = Path(remotedir).stem
                if parent != remotedir:
                    if not self.isdir(parent):
                        self.mkdir_p(parent, mode=mode)
                if stem:
                    self.mkdir(remotedir, mode=mode)
        except Exception as err:
            raise err

    def normalize(self, remotepath):
        '''Return the fully expanded path of a given location. This can be used
        to resolve symlinks or determine what the server believes to be the
        :attr:`.pwd`, by passing '.' as remotepath.

        :param str remotepath: Remote location to be normalized.

        :return: (str) Normalized path.

        :raises: IOError, if remotepath can't be resolved
        '''
        with self._sftp_channel() as channel:
            expanded_path = channel.normalize(drivedrop(remotepath))

        return expanded_path

    def open(self, remotefile, bufsize=-1, mode='r'):
        '''Open a file on the remote server.

        :param str remotefile: Path of remote file to open.
        :param str mode: *Default: read-only* - File access mode.
        :param int bufsize: *Default: -1* - Buffering in bytes.

        :returns: (obj) SFTPFile, a file-like object handler.

        :raises: IOError, if the file could not be opened.
        '''
        with self._sftp_channel(keepalive=True) as channel:
            remotefile = drivedrop(remotefile)
            flo = channel.open(remotefile, bufsize=bufsize, mode=mode)

        return flo

    def readlink(self, remotelink):
        '''Return the target of a symlink as an absolute path.

        :param str remotelink: Remote location of the symlink.

        :return: (str) Absolute path to target.
        '''
        with self._sftp_channel() as channel:
            remotelink = drivedrop(remotelink)
            link_destination = channel.normalize(channel.readlink(remotelink))

        return link_destination

    def remotetree(self, container, remotedir, localdir, recurse=True):
        '''Recursively map remote directory tree to a dictionary container.

        :param dict container: Hash table to save remote directory tree.
            {remotedir: [(remotedir/subdir, localdir/remotedir/subdir)]}
        :param str remotedir: Remote location to descend, use '.' to start at
            :attr:`.pwd`.
        :param str localdir: Location used as root of appended remote paths.
        :param bool recurse: *Default: True* - To recurse or not to recurse
            that is the question.

        :returns: None

        :raises: Exception
        '''
        try:
            localdir = Path(localdir).expanduser().as_posix()
            remotedir = self.normalize(remotedir)
            for attribute in self.listdir_attr(remotedir):
                if S_ISDIR(attribute.st_mode):
                    remote = Path(remotedir).joinpath(
                        attribute.filename).as_posix()
                    local = Path(localdir).joinpath(
                        Path(remote).stem).as_posix()
                    if remotedir in container.keys():
                        container[remotedir].append((remote, local))
                    else:
                        container[remotedir] = [(remote, local)]
                    if recurse:
                        self.remotetree(container, remote, local,
                                        recurse=recurse)
        except Exception as err:
            raise err

    def remove(self, remotefile):
        '''Delete the remote file. May include a path, if no path, then
        :attr:`.pwd` is used. This method only works on files.

        :param str remotefile: Remote file to delete.

        :returns: None

        :raises: IOError
        '''
        with self._sftp_channel() as channel:
            channel.remove(drivedrop(remotefile))

    def rename(self, remotepath, newpath):
        '''Rename a path on the remote host.

        :param str remotepath: Remote path to rename.

        :param str newpath: New name for remote path.

        :returns: None

        :raises: IOError
        '''
        with self._sftp_channel() as channel:
            channel.posix_rename(drivedrop(remotepath), drivedrop(newpath))

    def rmdir(self, remotedir):
        '''Delete remote directory.

        :param str remotedir: Remote directory to delete.

        :returns: None
        '''
        with self._sftp_channel() as channel:
            channel.rmdir(drivedrop(remotedir))

    def stat(self, remotepath):
        '''Return information about remote location.

        :param str remotepath: Remote location to stat.

        :returns: (obj) SFTPAttributes
        '''
        with self._sftp_channel() as channel:
            stat = channel.stat(drivedrop(remotepath))

        return stat

    def symlink(self, remote_src, remote_dest):
        '''Create a symlink for a remote file on the server

        :param str remote_src: path of original file
        :param str remote_dest: path of the created symlink

        :returns: None

        :raises: any underlying error, IOError if remote_dest already exists
        '''
        with self._sftp_channel() as channel:
            channel.symlink(remote_src, drivedrop(remote_dest))

    def truncate(self, remotepath, size):
        '''Change the size of the file specified by path. Used to modify the
        size of the file, just like the truncate method on Python file objects.
        The new file size is confirmed and returned.

        :param str remotepath: remote file path to modify
        :param int|long size: the new file size

        :returns: (int) new size of file

        :raises: IOError, if file does not exist
        '''
        with self._sftp_channel() as channel:
            remotepath = drivedrop(remotepath)
            channel.truncate(remotepath, size)
            size = channel.stat(remotepath).st_size

        return size

    @property
    def active_ciphers(self):
        '''Get tuple of currently used local and remote ciphers.

        :returns:
            (tuple of  str) currently used ciphers (local_cipher,
            remote_cipher)
        '''
        return self._transport.local_cipher, self._transport.remote_cipher

    @property
    def active_compression(self):
        '''Get tuple of local and remote compression status.

        :returns: (tuple of str) Compression status.
            (local_compression, remote_compression)
        '''
        local_compression = self._transport.local_compression
        remote_compression = self._transport.remote_compression

        return local_compression, remote_compression

    @property
    def logfile(self):
        '''Return logging setting.

        :returns: (str) logfile or (bool) False
        '''
        return self._cnopts.log

    @property
    def pwd(self):
        '''Return the current working directory.

        :returns: (str) Current working directory.
        '''
        with self._sftp_channel() as channel:
            pwd = channel.normalize('.')

        return pwd

    @property
    def remote_server_key(self):
        '''Return the remote server's key'''
        return self._transport.get_remote_server_key()

    @property
    def security_options(self):
        '''Return the transport security options.

        :returns: (obj) Security preferences for the underlying transport.
            These are tuples of acceptable `.ciphers`, `.digests`, `.key_types`
            and key exchange algorithms `.kex`, listed in order of preference.
        '''
        return self._transport.get_security_options()

    @property
    def sftp_client(self):
        '''Provide access to the underlying SFTPClient object. Client is not
        handled by context manager. Connection is closed with underlying
        transport if not done explicitly.

        :params: None

        :returns: (obj) Active SFTPClient object.
        '''
        with self._sftp_channel(keepalive=True) as channel:
            return channel

    @property
    def timeout(self):
        '''Get or set the underlying socket timeout for pending IO operations.

        :returns: (float|None) Seconds to wait for pending read/write operation
            before raising socket.timeout, or None for no timeout
        '''
        with self._sftp_channel() as channel:
            _channel = channel.get_channel()
            timeout = _channel.gettimeout()

        return timeout

    @timeout.setter
    def timeout(self, val):
        '''Setter for timeout'''
        self._timeout = val

    def __del__(self):
        '''Attempt to garbage collect if not explicitly closed.'''
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        '''GTFO'''
        self.close()



'''test sftpretty.chown'''

import pytest

from common import SKIP_IF_WIN, tempfile_containing
from pathlib import Path


@SKIP_IF_WIN  # uid comes through as 0, lacks support
def test_chown_uid(lsftp):
    '''test changing just the uid'''
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        org_attrs = lsftp.put(fname)
        uid = org_attrs.st_uid
        lsftp.chown(base_fname, uid=uid)
        new_attrs = lsftp.stat(base_fname)
        lsftp.remove(base_fname)
    assert new_attrs.st_gid == org_attrs.st_gid
    assert new_attrs.st_uid == uid


@SKIP_IF_WIN  # gid comes through as 0, lacks support
def test_chown_gid(lsftp):
    '''test changing just the gid'''
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        org_attrs = lsftp.put(fname)
        gid = org_attrs.st_gid
        lsftp.chown(base_fname, gid=gid)
        new_attrs = lsftp.stat(base_fname)
        lsftp.remove(base_fname)
    assert new_attrs.st_gid == gid
    assert new_attrs.st_uid == org_attrs.st_uid


def test_chown_none(lsftp):
    '''call .chown with no gid or uid specified'''
    with tempfile_containing() as fname:
        base_fname = Path(fname).name
        org_attrs = lsftp.put(fname)
        lsftp.chown(base_fname)
        new_attrs = lsftp.stat(base_fname)
        lsftp.remove(base_fname)
    assert new_attrs.st_gid == org_attrs.st_gid
    assert new_attrs.st_uid == org_attrs.st_uid


def test_chown_not_exist(lsftp):
    '''call .chown on a non-existing path'''
    with pytest.raises(IOError):
        lsftp.chown('i-do-not-exist.txt', 666)


# TODO
# def test_chown_ro_server(psftp):
#     '''call .chown against path on read-only server'''
#     with pytest.raises(IOError):
#         psftp.chown('readme.txt', gid=1000, uid=1000)
